#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device classes for a CryLas pulsed laser controller and a CryLas laser attenuator,
using serial communication.

There are three modes of operation for the laser
1. Laser-internal hardware trigger (default): fixed to 20 Hz and max energy per pulse.
2. Laser-internal software trigger (for diagnosis only).
3. External trigger: required for arbitrary pulse energy or repetition rate. Switch to
"external" on the front panel of laser controller for using option 3.

After switching on the laser with laser_on(), the system must stabilize
for some minutes. Do not apply abrupt changes of pulse energy or repetition rate.

Manufacturer homepage:
https://www.crylas.de/products/pulsed_laser.html

TODO: Fix typing
"""

import logging
import math
import re
import time
from collections.abc import Callable
from datetime import datetime
from threading import Event

from aenum import Enum, IntEnum

from hvl_ccb.comm.serial import (
    SerialCommunication,
    SerialCommunicationBytesize,
    SerialCommunicationConfig,
    SerialCommunicationParity,
    SerialCommunicationStopbits,
)
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError, SingleCommDevice
from hvl_ccb.utils.poller import Poller
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


@configdataclass
class CryLasLaserSerialCommunicationConfig(SerialCommunicationConfig):
    #: Baudrate for CryLas laser is 19200 baud
    baudrate: int = 19200

    #: CryLas laser does not use parity
    parity: str | SerialCommunicationParity = SerialCommunicationParity.NONE

    #: CryLas laser uses one stop bit
    stopbits: int | SerialCommunicationStopbits = SerialCommunicationStopbits.ONE

    #: One byte is eight bits long
    bytesize: int | SerialCommunicationBytesize = SerialCommunicationBytesize.EIGHTBITS

    #: The terminator is LF
    terminator: bytes = b"\n"

    #: use 10 seconds timeout as default (a long timeout is needed!)
    timeout: Number = 10


class CryLasLaserSerialCommunication(SerialCommunication):
    """
    Specific communication protocol implementation for the CryLas laser controller.
    Already predefines device-specific protocol parameters in config.
    """

    @staticmethod
    def config_cls():
        return CryLasLaserSerialCommunicationConfig

    READ_TEXT_SKIP_PREFIXES = (">", "MODE:")
    """Prefixes of lines that are skipped when read from the serial port."""

    def read(
        self,
    ) -> str:
        """
        Read first line of text from the serial port that does not start with any of
        `self.READ_TEXT_SKIP_PREFIXES`.

        :return: String read from the serial port; `''` if there was nothing to read.
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        with self.access_lock:
            line = super().read()
            while line.startswith(self.READ_TEXT_SKIP_PREFIXES):
                logger.debug(f'Laser com: "{line.rstrip()}" SKIPPED')
                line = super().read_text()
            logger.debug(f'Laser com: "{line.rstrip()}"')
            return line

    def query(self, cmd: str, prefix: str, post_cmd: str | None = None) -> str:
        """
        Send a command, then read the com until a line starting with prefix,
        or an empty line, is found. Returns the line in question.

        :param cmd: query message to send to the device
        :param prefix: start of the line to look for in the device answer
        :param post_cmd: optional additional command to send after the query
        :return: line in question as a string
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        # TODO: Make use of SyncCommunicationProtocol  # noqa: FIX002
        # https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/372

        with self.access_lock:
            # send the command
            self.write(cmd)
            # read the com until the prefix or an empty line is found
            line = self.read().strip()
            while line and not line.startswith(prefix):
                line = self.read().strip()
            if post_cmd is not None:
                # send an optional post-command message to send to the device
                self.write(post_cmd)
            return line

    def query_all(self, cmd: str, prefix: str) -> list:
        """
        Send a command, then read the com until a line starting with prefix,
        or an empty line, is found. Returns a list of successive lines starting with
        prefix.

        :param cmd: query message to send to the device
        :param prefix: start of the line to look for in the device answer
        :return: line in question as a string
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        # TODO: Make use of SyncCommunicationProtocol  # noqa: FIX002
        # https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/372

        with self.access_lock:
            # send the command
            self.write(cmd)
            # read the com until the prefix or an empty line is found
            line = self.read().strip()
            while line and not line.startswith(prefix):
                line = self.read().strip()
            answer = []
            while line.startswith(prefix):
                answer.append(line)
                line = self.read().strip()
            return answer


class CryLasLaserShutterStatus(Enum):
    """
    Status of the CryLas laser shutter
    """

    CLOSED = 0
    OPENED = 1


@configdataclass
class CryLasLaserConfig:
    """
    Device configuration dataclass for the CryLas laser controller.
    """

    # status of the shutter
    ShutterStatus = CryLasLaserShutterStatus

    # calibration factor for the pulse energy
    calibration_factor: Number = 4.35

    # polling period (s) to check back on the laser status if not ready
    # it should be longer than the communication timeout
    polling_period: int | float = 12

    # timeout (s) when polling the laser status, CryLasLaserError is raised on timeout
    polling_timeout: Number = 300

    # automatically turn on the laser when ready
    auto_laser_on: bool = True

    # status of the shutter to be set on start
    init_shutter_status: int | CryLasLaserShutterStatus = (
        CryLasLaserShutterStatus.CLOSED
    )

    def clean_values(self) -> None:
        if self.calibration_factor <= 0:
            msg = "The calibration factor should be positive."
            raise ValueError(msg)
        if self.polling_period <= 0:
            msg = "The polling period should be positive."
            raise ValueError(msg)
        if self.polling_timeout <= 0:
            msg = "The polling timeout should be positive."
            raise ValueError(msg)
        self.force_value(  # type: ignore[attr-defined]
            "init_shutter_status", self.ShutterStatus(self.init_shutter_status)
        )


class CryLasLaserPoller(Poller):
    """
    Poller class for polling the laser status until the laser is ready.

    :raises CryLasLaserError: if the timeout is reached before the laser is ready
    :raises SerialCommunicationIOError: when communication port is closed.
    """

    def __init__(  # noqa: ANN204
        self,
        spoll_handler: Callable,
        check_handler: Callable,
        check_laser_status_handler: Callable,
        polling_delay_sec: Number = 0,
        polling_interval_sec: Number = 1,
        polling_timeout_sec: Number | None = None,
    ):
        """
        Initialize the polling helper.

        :param spoll_handler: Polling function.
        :param check_handler: Check polling results.
        :param check_laser_status_handler: Check laser status.
        :param polling_delay_sec: Delay before starting the polling, in seconds.
        :param polling_interval_sec: Polling interval, in seconds.
        """
        super().__init__(
            spoll_handler, polling_delay_sec, polling_interval_sec, polling_timeout_sec
        )

        self._check_handler = check_handler
        self._check_laser_status_handler = check_laser_status_handler

    def _if_poll_again(
        self, stop_event: Event, delay_sec: Number, stop_time: Number | None
    ) -> bool:
        """
        Check if to poll again.

        :param stop_event: Polling stop event.
        :param delay_sec: Delay time (in seconds).
        :param stop_time: Absolute stop time.
        :return: `True` if another polling handler call is due, `False` otherwise.
        """
        if_poll_again = super()._if_poll_again(stop_event, delay_sec, stop_time)
        is_laser_ready = self._check_laser_status_handler()

        return if_poll_again and not is_laser_ready

    def _poll_until_stop_or_timeout(self, stop_event: Event) -> object | None:
        """
        Thread for polling until stopped, timed-out or the laser is ready.

        :param stop_event: Event used to stop the polling
        :return: Last result of the polling function call
        """
        last_result = super()._poll_until_stop_or_timeout(stop_event)

        if stop_event.wait(0):
            logger.info(f"[{datetime.now()}] Polling status: ... STOPPED")
            return last_result

        self._check_handler()
        return last_result


class CryLasLaser(SingleCommDevice):
    """
    CryLas laser controller device class.
    """

    class LaserStatus(Enum):
        """
        Status of the CryLas laser
        """

        UNREADY_INACTIVE = 0
        READY_INACTIVE = 1
        READY_ACTIVE = 2

        @property
        def is_ready(self) -> bool:
            return self is not CryLasLaser.LaserStatus.UNREADY_INACTIVE

        @property
        def is_inactive(self) -> bool:
            return self is not CryLasLaser.LaserStatus.READY_ACTIVE

    # status of the shutter
    ShutterStatus = CryLasLaserShutterStatus

    class AnswersShutter(Enum):
        """
        Standard answers of the CryLas laser controller to `'Shutter'` command passed
        via `com`.
        """

        OPENED = "Shutter aktiv"
        CLOSED = "Shutter inaktiv"

    class AnswersStatus(Enum):
        """
        Standard answers of the CryLas laser controller to `'STATUS'` command passed
        via `com`.
        """

        TEC1 = "STATUS: TEC1 Regulation ok"
        TEC2 = "STATUS: TEC2 Regulation ok"
        HEAD = "STATUS: Head ok"
        READY = "STATUS: System ready"
        ACTIVE = "STATUS: Laser active"
        INACTIVE = "STATUS: Laser inactive"

    class RepetitionRates(IntEnum, init="value send_value"):  # type: ignore[call-arg]
        """
        Repetition rates for the internal software trigger in Hz
        """

        HARDWARE = 0, 0  # software trigger is disabled, hardware trigger is used
        # instead (hardware trigger may be internal or external).
        SOFTWARE_INTERNAL_TEN = 10, 1
        SOFTWARE_INTERNAL_TWENTY = 20, 2
        SOFTWARE_INTERNAL_SIXTY = 60, 3

    def __init__(self, com, dev_config=None):  # noqa: ANN204
        # Call superclass constructor
        super().__init__(com, dev_config)

        # laser status
        self.laser_status = self.LaserStatus.UNREADY_INACTIVE

        # shutter status
        self.shutter_status = None

        # command repetition rate
        self.repetition_rate = None

        # command pulse energy (micro joule)
        self._target_pulse_energy = None

        # thread that polls the laser status until it is ready
        self._status_poller = None

    @property
    def target_pulse_energy(self) -> int | None:
        return self._target_pulse_energy

    @staticmethod
    def default_com_cls():
        return CryLasLaserSerialCommunication

    @staticmethod
    def config_cls():
        return CryLasLaserConfig

    def start(self) -> None:
        """
        Opens the communication protocol and configures the device.

        :raises SerialCommunicationIOError: when communication port cannot be opened
        """

        logger.info(f"Starting device {self}")

        # open the com
        super().start()

        # set the init shutter status
        self.set_init_shutter_status()
        # check if the laser is ready to be turned on
        self.update_laser_status()
        if not self.laser_status.is_ready:
            logger.info("Laser not ready yet.")
            # optionally, block execution until laser is ready
            if self.config.auto_laser_on and not self._is_polling():
                self._start_polling()
        elif self.config.auto_laser_on:
            # turn on the laser
            self.laser_on()

    def _start_polling(self):
        """
        Start polling laser status.
        """
        logger.info("Start polling laser status.")
        self._status_poller = CryLasLaserPoller(
            spoll_handler=self.update_laser_status,
            check_handler=self._after_polling_check_status,
            check_laser_status_handler=self._check_laser_status_handler,
            polling_delay_sec=self.config.polling_period,
            polling_interval_sec=self.config.polling_period,
            polling_timeout_sec=self.config.polling_timeout,
        )
        self._status_poller.start_polling()

    def _check_laser_status_handler(self) -> bool:
        """
        Checks whether the laser status is ready.
        """
        return self.laser_status.is_ready

    def _after_polling_check_status(self) -> None:
        """
        Thread for polling the laser status until the laser is ready.

        :raises CryLasLaserError: if the timeout is reached before the laser is ready
        :raises SerialCommunicationIOError: when communication port is closed.
        """

        if not self.laser_status.is_ready:
            err_msg = "Laser is not yet ready but status polling timed out."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        logger.info(f"[{datetime.now()}] Polling status: ... DONE")
        if self.config.auto_laser_on:
            self.laser_on()

    def _wait_for_polling_result(self):
        """
        Wait for until polling function returns a result as well as any exception that
        might have been raised within a thread.

        :return: polling function result
        :raises: polling function errors
        """
        return self._status_poller.wait_for_polling_result()

    def _stop_polling(self):
        """
        Stop polling laser status.

        :return: polling function result
        :raises: polling function errors
        """
        # use the event to stop the thread
        logger.info("Stop polling laser status.")
        self._status_poller.stop_polling()
        return self._wait_for_polling_result()

    def _is_polling(self) -> bool:
        """
        Check if device status is being polled.

        :return: `True` when polling thread is set and alive
        """
        return self._status_poller is not None and self._status_poller.is_polling()

    def wait_until_ready(self) -> None:
        """
        Block execution until the laser is ready

        :raises CryLasLaserError: if the polling thread stops before the laser is ready
        """

        if not self.laser_status.is_ready:
            if not self._is_polling():
                self._start_polling()
            logger.info("Waiting until the laser is ready...")
            self._wait_for_polling_result()
        else:
            logger.info("No need waiting, the laser is already ready.")

    def stop(self) -> None:
        """
        Stops the device and closes the communication protocol.

        :raises SerialCommunicationIOError: if com port is closed unexpectedly
        :raises CryLasLaserError: if laser_off() or close_shutter() fail
        """

        if self.com.is_open:
            # turn off the laser
            self.laser_off()
            # close the laser shutter
            self.close_shutter()
            # cancel the polling thread in case still running
            if self._is_polling():
                self._stop_polling()
                logger.info("The laser polling thread was stopped.")
            # close the com
            super().stop()
        else:
            logger.warning("Could not turn off the laser, com was already closed.")

    def update_laser_status(self) -> None:
        """
        Update the laser status to `LaserStatus.NOT_READY` or `LaserStatus.INACTIVE` or
        `LaserStatus.ACTIVE`.

        Note: laser never explicitly says that it is not ready  (
        `LaserStatus.NOT_READY`) in response to `'STATUS'` command. It only says
        that it is ready (heated-up and implicitly inactive/off) or active (on). If
        it's not either of these then the answer is `Answers.HEAD`. Moreover,
        the only time the laser explicitly says that its status is inactive (
        `Answers.INACTIVE`) is after issuing a 'LASER OFF' command.

        :raises SerialCommunicationIOError: when communication port is not opened
        """

        # query the status
        answer_list = self.com.query_all("STATUS", "STATUS:")
        # analyze the answer
        if not answer_list:
            err_msg = "Command to query the laser status did not return an answer."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        # at least one line in answer_list
        for line in answer_list:
            answer = self.AnswersStatus(line)
            if answer in (
                self.AnswersStatus.TEC1,
                self.AnswersStatus.TEC2,
                self.AnswersStatus.HEAD,
            ):
                new_status = self.LaserStatus.UNREADY_INACTIVE
                info_msg = "The laser is not ready."
            elif answer is self.AnswersStatus.READY:
                new_status = self.LaserStatus.READY_INACTIVE
                info_msg = "The laser is ready."
            elif answer is self.AnswersStatus.ACTIVE:
                new_status = self.LaserStatus.READY_ACTIVE
                info_msg = "The laser is on."
            else:
                err_msg = f'Unexpected "STATUS" query response {answer.value}.'
                logger.error(err_msg)
                raise CryLasLaserError(err_msg)
        self.laser_status = new_status
        logger.info(info_msg)

    def update_shutter_status(self) -> None:
        """
        Update the shutter status (OPENED or CLOSED)

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # query the status
        answer = self.com.query("Shutter", "Shutter")
        # analyse the answer
        if not answer:
            err_msg = "Command to query the shutter status did not return an answer."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        answer = self.AnswersShutter(answer)
        if answer is self.AnswersShutter.CLOSED:
            logger.info("The laser shutter is currently CLOSED.")
            self.shutter_status = self.ShutterStatus.CLOSED
        elif answer is self.AnswersShutter.OPENED:
            logger.info("The laser shutter is currently OPENED.")
            self.shutter_status = self.ShutterStatus.OPENED

    def update_repetition_rate(self) -> None:
        """
        Query the laser repetition rate.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # query the repetition rate
        answer = self.com.query("BOO IP", "Impuls=")
        # analyse the answer
        if not answer:
            err_msg = "Querying the repetition rate did not return an answer."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)
        if answer == "Impuls=disabled, extern Trigger":
            logger.info(
                "The laser is using a hardware trigger (may be internal or external)."
            )
            self.repetition_rate = self.RepetitionRates.HARDWARE
        elif answer.startswith("Impuls=enabled"):
            match = re.search(r"\d+", answer)
            if match is None:
                err_msg = (
                    "Expected rate integer to follow 'Impuls=enabled' answer "
                    f"in {answer} while querying the repetition."
                )
                logger.error(err_msg)
                raise CryLasLaserError(err_msg)
            rate = int(match.group(0))
            self.repetition_rate = self.RepetitionRates(rate)
            logger.info(
                f"The laser is using the internal software trigger at {rate} Hz."
            )

    def update_target_pulse_energy(self) -> None:
        """
        Query the laser pulse energy.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # query the repetition rate
        answer = self.com.query("BOO SE", "PD-Sollwert=")
        # analyse the answer
        if not answer:
            err_msg = (
                "Command to query the target pulse energy did not return an answer."
            )
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        current_value = int(answer.split("=")[1])
        current_energy = int(current_value * self.config.calibration_factor / 1000)
        # update the pulse energy attribute
        self._target_pulse_energy = current_energy
        logger.info(f"The target pulse energy is {current_energy} uJ.")

    def laser_on(self) -> None:
        """
        Turn the laser on.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserNotReadyError: if the laser is not ready to be turned on
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        if not self.laser_status.is_ready:
            msg = "Laser not ready, cannot be turned on yet."
            raise CryLasLaserNotReadyError(msg)

        # send the command
        answer = self.com.query("LASER ON", "STATUS: Laser")
        # analyse the answer
        if not answer or self.AnswersStatus(answer) != self.AnswersStatus.ACTIVE:
            if not self.laser_status.is_inactive:
                logger.info("Laser is already on.")
            else:
                err_msg = (
                    "Command to turn on the laser "
                    f"{'failed' if answer else 'did not return an answer'}."
                )
                logger.error(err_msg)
                raise CryLasLaserError(err_msg)
        else:
            self.laser_status = self.LaserStatus.READY_ACTIVE
            logger.info("Laser is turned on.")

    def laser_off(self) -> None:
        """
        Turn the laser off.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # send the command
        answer = self.com.query("LASER OFF", "STATUS: Laser")
        # analyse the answer
        if not answer or self.AnswersStatus(answer) != self.AnswersStatus.INACTIVE:
            if self.laser_status.is_inactive:
                logger.info("Laser is already off.")
            else:
                err_msg = (
                    "Command to turn off the laser "
                    f"{'failed' if answer else 'did not return an answer'}."
                )
                logger.error(err_msg)
                raise CryLasLaserError(err_msg)
        else:
            self.laser_status = self.LaserStatus.READY_INACTIVE
            logger.info("Laser is turned off.")

    def open_shutter(self) -> None:
        """
        Open the laser shutter.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # send the command
        answer = self.com.query(
            f"Shutter {self.config.ShutterStatus.OPENED.value}", "Shutter"
        )
        # analyse the answer
        if not answer or self.AnswersShutter(answer) != self.AnswersShutter.OPENED:
            err_msg = (
                "Opening laser shutter "
                f"{'failed' if answer else 'did not return an answer'}."
            )
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        logger.info("Opening laser shutter succeeded.")
        self.shutter_status = self.ShutterStatus.OPENED

    def close_shutter(self) -> None:
        """
        Close the laser shutter.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # send the command
        answer = self.com.query(
            f"Shutter {self.config.ShutterStatus.CLOSED.value}", "Shutter"
        )
        # analyse the answer
        if not answer or self.AnswersShutter(answer) != self.AnswersShutter.CLOSED:
            err_msg = (
                "Closing laser shutter "
                f"{'failed' if answer else 'did not return an answer'}."
            )
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        logger.info("Closing laser shutter succeeded.")
        self.shutter_status = self.ShutterStatus.CLOSED

    def set_init_shutter_status(self) -> None:
        """
        Open or close the shutter, to match the configured shutter_status.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        self.update_shutter_status()
        if self.config.init_shutter_status != self.shutter_status:
            if self.config.init_shutter_status == self.ShutterStatus.CLOSED:
                self.close_shutter()
            elif self.config.init_shutter_status == self.ShutterStatus.OPENED:
                self.open_shutter()

    def get_pulse_energy_and_rate(self) -> tuple[int, int]:
        """
        Use the debug mode, return the measured pulse energy and rate.

        :return: (energy in micro joule, rate in Hz)
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if the device does not answer the query
        """

        # send the command (enter the debug mode)
        answer = self.com.query("DB1", "IST:", "DB0")
        if not answer:
            err_msg = "Command to find laser energy and rate did not return an answer."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        # the answer should look like 'IST: LDTemp=x, NLOTemp=x, CaseTemp=x,
        # PD=x, D_I= x, freq=x, LDE=x, RegOFF=x' where x are integers
        # find the values of interest
        values = re.findall(r"\d+", answer)
        energy = int(int(values[3]) * self.config.calibration_factor / 1000)
        rate = int(values[5])
        logger.info(f"Laser energy: {energy} uJ, repetition rate: {rate} Hz.")
        return energy, rate

    def set_repetition_rate(self, rate: int | RepetitionRates) -> None:
        """
        Sets the repetition rate of the internal software trigger.

        :param rate: frequency (Hz) as an integer
        :raises ValueError: if rate is not an accepted value in RepetitionRates Enum
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if success is not confirmed by the device
        """

        # check if the value is ok
        if not isinstance(rate, self.RepetitionRates):
            try:
                rate = self.RepetitionRates(rate)
            except ValueError as e:
                logger.exception("Error of Crylas: Wrong RepetitionRate", exc_info=e)
                raise
        # send the corresponding value to the controller
        answer = self.com.query(f"BOO IP{rate.send_value}", "Impuls=")
        # check the success of the command
        if rate != self.RepetitionRates.HARDWARE:
            # check whether the expected answer is obtained
            if answer != f"Impuls=enabled {rate}Hz":
                err_msg = (
                    "Setting the repetition rate failed. Controller answered:"
                    f" {answer}."
                )
                logger.error(err_msg)
                raise CryLasLaserError(err_msg)

            logger.info(f"Laser internal software trigger set to {rate.value} Hz")
        else:
            # For the internal hardware trigger rate, there is no way to check success
            logger.info("Using laser internal hardware trigger.")
        # update the repetition rate attribute
        self.repetition_rate = rate

    def set_pulse_energy(self, energy: int) -> None:
        """
        Sets the energy of pulses (works only with external hardware trigger).
        Proceed with small energy steps, or the regulation may fail.

        :param energy: energy in micro joule
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasLaserError: if the device does not confirm success
        """

        # convert the input value into the required command value
        command_value = int(energy * 1000 / self.config.calibration_factor)
        # send the command
        answer = self.com.query(f"BOO SE {command_value}", "PD-Sollwert=")
        # analyse the answer
        if not answer:
            err_msg = "Command to set the pulse energy did not return an answer."
            logger.error(err_msg)
            raise CryLasLaserError(err_msg)

        current_value = int(answer.split("=")[1])
        current_energy = int(current_value * self.config.calibration_factor / 1000)
        # update the pulse energy attribute
        self._target_pulse_energy = current_energy
        if current_value != command_value:
            message = (
                "Command to set the pulse energy failed. "
                f"The pulse energy is currently set to {current_energy} "
                "micro joule. Setting a different value is only possible "
                "when using an external hardware trigger."
            )
            logger.error(message)
            raise CryLasLaserError(message)

        logger.info(f"Successfully set pulse energy to {energy}")


class CryLasLaserError(DeviceError):
    """
    General error with the CryLas Laser.
    """


class CryLasLaserNotReadyError(CryLasLaserError):
    """
    Error when trying to turn on the CryLas Laser before it is ready.
    """


@configdataclass
class CryLasAttenuatorSerialCommunicationConfig(SerialCommunicationConfig):
    #: Baudrate for CryLas attenuator is 9600 baud
    baudrate: int = 9600

    #: CryLas attenuator does not use parity
    parity: str | SerialCommunicationParity = SerialCommunicationParity.NONE

    #: CryLas attenuator uses one stop bit
    stopbits: int | SerialCommunicationStopbits = SerialCommunicationStopbits.ONE

    #: One byte is eight bits long
    bytesize: int | SerialCommunicationBytesize = SerialCommunicationBytesize.EIGHTBITS

    #: No terminator
    terminator: bytes = b""

    #: use 3 seconds timeout as default
    timeout: Number = 3


class CryLasAttenuatorSerialCommunication(SerialCommunication):
    """
    Specific communication protocol implementation for
    the CryLas attenuator.
    Already predefines device-specific protocol parameters in config.
    """

    @staticmethod
    def config_cls():
        return CryLasAttenuatorSerialCommunicationConfig


@configdataclass
class CryLasAttenuatorConfig:
    """
    Device configuration dataclass for CryLas attenuator.
    """

    # initial/default attenuation value which is set on start()
    init_attenuation: Number = 0
    response_sleep_time: Number = 1

    def clean_values(self) -> None:
        if not 0 <= self.init_attenuation <= 100:
            msg = "Attenuation should be between 0 and 100 included."
            raise ValueError(msg)
        if self.response_sleep_time <= 0:
            msg = "Response sleep time should be positive."
            raise ValueError(msg)


class CryLasAttenuator(SingleCommDevice):
    """
    Device class for the CryLas laser attenuator.
    """

    def __init__(self, com, dev_config=None):  # noqa: ANN204
        # Call superclass constructor
        super().__init__(com, dev_config)

        # attenuation of the laser light in percent (not determined yet)
        self._attenuation = None

    @staticmethod
    def default_com_cls():
        return CryLasAttenuatorSerialCommunication

    @staticmethod
    def config_cls():
        return CryLasAttenuatorConfig

    @property
    def attenuation(self) -> Number:
        return self._attenuation

    @property
    def transmission(self) -> Number:
        return 100 - self._attenuation

    def start(self) -> None:
        """
        Open the com, apply the config value 'init_attenuation'

        :raises SerialCommunicationIOError: when communication port cannot be opened
        """

        super().start()
        self.set_init_attenuation()

    def set_init_attenuation(self) -> None:
        """
        Sets the attenuation to its configured initial/default value

        :raises SerialCommunicationIOError: when communication port is not opened
        """

        self.set_attenuation(self.config.init_attenuation)

    def set_attenuation(self, percent: Number) -> None:
        """
        Set the percentage of attenuated light (inverse of set_transmission).
        :param percent: percentage of attenuation, number between 0 and 100
        :raises ValueError: if param percent not between 0 and 100
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasAttenuatorError: if the device does not confirm success
        """

        if not 0 <= percent <= 100:
            msg = "Attenuation should be between 0 and 100 included."
            raise ValueError(msg)

        pulse = int(math.asin((50 - percent) * 0.02) * 536.5 + 6000)
        prepulse = pulse - 400
        # prepare the values to send with the motor protocol
        lsb = prepulse % 128
        msb = math.floor(prepulse / 128)
        # send the values
        self.com.write_bytes(bytes([132, 0, lsb, msb]))
        time.sleep(self.config.response_sleep_time)
        lsb = pulse % 128
        msb = math.floor(pulse / 128)
        self.com.write_bytes(bytes([132, 0, lsb, msb]))
        time.sleep(self.config.response_sleep_time / 10)
        self.com.write_bytes(bytes([161]))
        time.sleep(self.config.response_sleep_time / 10)
        b1 = self.com.read_single_bytes()
        b2 = self.com.read_single_bytes()
        logger.debug(f"b1 = {b1}, b2 = {b2}")
        if b1 != b"\x00" or b2 != b"\x00":
            err_msg = f"Setting laser attenuation to {percent} percents failed."
            logger.error(err_msg)
            raise CryLasAttenuatorError(err_msg)

        logger.info(f"Successfully set laser attenuation to {percent} Cpercents.")
        self._attenuation = percent

    def set_transmission(self, percent: Number) -> None:
        """
        Set the percentage of transmitted light (inverse of set_attenuation).
        :param percent: percentage of transmitted light
        :raises ValueError: if param percent not between 0 and 100
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises CryLasAttenuatorError: if the device does not confirm success
        """

        self.set_attenuation(100 - percent)


class CryLasAttenuatorError(DeviceError):
    """
    General error with the CryLas Attenuator.
    """



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the SST Luminox Oxygen sensor device class.
"""

import pytest

from hvl_ccb.comm.serial import SerialCommunicationIOError
from hvl_ccb.dev.sst_luminox import (
    Luminox,
    LuminoxConfig,
    LuminoxMeasurementType,
    LuminoxMeasurementTypeError,
    LuminoxOutputMode,
    LuminoxOutputModeError,
)
from masked_comm.serial import LuminoxLoopSerialCommunication


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "timeout": 0.05,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "wait_sec_post_activate": 0.01,
        "wait_sec_trials_activate": 0.01,
        "nr_trials_activate": 1,
    }


@pytest.fixture
def testdev(com_config, dev_config):
    serial_port = LuminoxLoopSerialCommunication(com_config)
    with Luminox(serial_port, dev_config) as lumi:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, lumi


@pytest.mark.parametrize(
    ("config_key", "wrong_value"),
    [
        ("wait_sec_post_activate", 0),
        ("wait_sec_post_activate", -1),
        ("wait_sec_trials_activate", 0),
        ("wait_sec_trials_activate", -1),
        ("nr_trials_activate", 0),
    ],
)
def test_invalid_config_value(dev_config, config_key, wrong_value) -> None:
    invalid_config = dict(dev_config)
    invalid_config[config_key] = wrong_value
    with pytest.raises(ValueError):
        LuminoxConfig(**invalid_config)


def test_instantiation(com_config, dev_config) -> None:
    mbw = Luminox(com_config, dev_config)
    assert mbw is not None


def test_com_error(com_config, dev_config) -> None:
    wrong_config = dict(com_config)
    wrong_config["port"] = "NOT A PORT"
    lumi = Luminox(wrong_config, dev_config)
    assert not lumi.com.is_open

    with pytest.raises(SerialCommunicationIOError):
        lumi.start()

    lumi = Luminox(com_config, dev_config)
    assert not lumi.com.is_open


def test_activate_output(testdev) -> None:
    com, lumi = testdev
    com.put_text(f"M 00{com.config.terminator_str()}")
    lumi.activate_output(LuminoxOutputMode.streaming)
    assert lumi.output == LuminoxOutputMode.streaming

    with pytest.raises(LuminoxOutputModeError):
        lumi.activate_output(LuminoxOutputMode.polling)


def test_start(testdev) -> None:
    _com, lumi = testdev

    # starting again should work
    lumi.start()


def test_read_stream(testdev) -> None:
    com, lumi = testdev
    expected_measurement_values = {
        LuminoxMeasurementType.partial_pressure_o2: "0190.1",
        LuminoxMeasurementType.temperature_sensor: "-27.3",
        LuminoxMeasurementType.barometric_pressure: "0970",
        LuminoxMeasurementType.percent_o2: "020.31",
        LuminoxMeasurementType.sensor_status: "0000",
    }
    read_txt_stream = (
        " ".join(
            f"{measurement} {value}"
            for (measurement, value) in expected_measurement_values.items()
        )
        + com.config.terminator_str()
    )
    com.put_text(read_txt_stream)

    lumi.output = LuminoxOutputMode.polling
    with pytest.raises(LuminoxOutputModeError):
        lumi.read_streaming()

    lumi.output = LuminoxOutputMode.streaming
    readout = lumi.read_streaming()
    expected_measurement_values = {
        measurement: float(value)
        for (measurement, value) in expected_measurement_values.items()
    }
    assert readout == expected_measurement_values


@pytest.mark.parametrize(
    "wrong_stream_read",
    [
        "O 0190.1 T -27.3 P  e 0000 % 020.31",  # missing value
        "O 0190.1 T -27.3 P 0970 e 0000 % ",  # missing value #2
        "O 0190.1 T -27.3 0970 e 0000 % 020.31",  # missing type symbol
    ],
)
def test_read_stream_measurement_type_error(testdev, wrong_stream_read) -> None:
    com, lumi = testdev
    lumi.output = LuminoxOutputMode.streaming
    com.put_text(wrong_stream_read)
    with pytest.raises(LuminoxMeasurementTypeError):
        lumi.read_streaming()


@pytest.mark.parametrize(
    ("measurement_type", "value_str"),
    [
        (LuminoxMeasurementType.partial_pressure_o2, f"{190.1:06.1f}"),
        (LuminoxMeasurementType.temperature_sensor, f"{-12.3:05.1f}"),
        (LuminoxMeasurementType.serial_number, "00123 56895"),
    ],
)
def test_query_single_measurement(testdev, measurement_type, value_str) -> None:
    com, lumi = testdev

    com.put_text(f"{measurement_type.command} {value_str}{com.config.terminator_str()}")

    lumi.output = LuminoxOutputMode.polling
    value = lumi.query_polling(measurement_type)
    expected_value = measurement_type.cast_type(value_str)
    assert value == expected_value


def test_query_single_measurement_all_measurements(testdev) -> None:
    com, lumi = testdev

    expected_measurement_values = {
        LuminoxMeasurementType.partial_pressure_o2: "0190.1",
        LuminoxMeasurementType.temperature_sensor: "-27.3",
        LuminoxMeasurementType.barometric_pressure: "0970",
        LuminoxMeasurementType.percent_o2: "020.31",
        LuminoxMeasurementType.sensor_status: "0000",
    }
    read_txt_stream = (
        " ".join(
            f"{measurement} {value}"
            for (measurement, value) in expected_measurement_values.items()
        )
        + com.config.terminator_str()
    )
    com.put_text(read_txt_stream)

    lumi.output = LuminoxOutputMode.polling
    value = lumi.query_polling(LuminoxMeasurementType.all_measurements)
    expected_measurement_values = {
        measurement: float(value)
        for (measurement, value) in expected_measurement_values.items()
    }
    assert value == expected_measurement_values


def test_query_single_measurement_error_streaming_mode(testdev) -> None:
    com, lumi = testdev

    measurement_type = LuminoxMeasurementType.partial_pressure_o2
    expected_value = 190.1
    com.put_text(
        f"{measurement_type} {expected_value:06.1f}{com.config.terminator_str()}"
    )

    lumi.output = LuminoxOutputMode.streaming
    with pytest.raises(LuminoxOutputModeError):
        lumi.query_polling(LuminoxMeasurementType.partial_pressure_o2)


def test_query_single_measurement_warning(testdev) -> None:
    com, lumi = testdev

    expected_value = 190.1
    com.put_text(
        f"{LuminoxMeasurementType.partial_pressure_o2} {expected_value:06.1f}"
        f" IGNORE ME{com.config.terminator_str()}"
    )

    lumi.output = LuminoxOutputMode.polling
    value = lumi.query_polling(LuminoxMeasurementType.partial_pressure_o2)
    assert value == expected_value
    # opt: check if the warning was actually issued


@pytest.mark.parametrize(
    "wrong_partial_pressure_o2_read",
    [
        "Z 0190.1",  # wrong type symbol
        str(LuminoxMeasurementType.partial_pressure_o2),  # missing value
        "0190.1",  # missing type symbol
    ],
)
def test_query_single_measurement_type_error(
    testdev, wrong_partial_pressure_o2_read
) -> None:
    com, lumi = testdev
    lumi.output = LuminoxOutputMode.polling
    com.put_text(wrong_partial_pressure_o2_read)
    with pytest.raises(LuminoxMeasurementTypeError):
        lumi.query_polling(LuminoxMeasurementType.partial_pressure_o2)


def test_query_single_measurement_str_enum_name(testdev) -> None:
    com, lumi = testdev

    measurement_type = LuminoxMeasurementType.serial_number
    expected_value = "00123 56895"
    com.put_text(
        f"{measurement_type.command} {expected_value}{com.config.terminator_str()}"
    )

    lumi.output = LuminoxOutputMode.polling
    value = lumi.query_polling(measurement_type.name)
    assert value == expected_value

    with pytest.raises(ValueError):
        lumi.query_polling(measurement_type.name.replace("_", " "))


def test_query_single_measurement_str_enum_value(testdev) -> None:
    com, lumi = testdev

    measurement_type = LuminoxMeasurementType.serial_number
    expected_value = "00123 56895"
    com.put_text(
        f"{measurement_type.command} {expected_value}{com.config.terminator_str()}"
    )

    lumi.output = LuminoxOutputMode.polling
    value = lumi.query_polling(measurement_type.value)
    assert value == expected_value

    with pytest.raises(ValueError):
        lumi.query_polling(measurement_type.command + " ?")



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Errors and their codes of FuG
"""

import logging

from hvl_ccb.dev.base import DeviceError
from hvl_ccb.utils.enum import NameEnum

logger = logging.getLogger(__name__)


class FuGError(DeviceError):
    """
    Error with the FuG voltage source.
    """

    def __init__(self, *args, **kwargs) -> None:
        self.errorcode: str = kwargs.pop("errorcode", "")
        """
        Errorcode from the Probus, see documentation of Probus V chapter 5.
        Errors with three-digit errorcodes are thrown by this python module.
        """
        super().__init__(*args, **kwargs)


class FuGErrorcodes(NameEnum, init="description possible_reason"):  # type: ignore[call-arg]
    """
    The power supply can return an errorcode. These errorcodes are handled by this
    class. The original errorcodes from the source are with one or two digits,
    see documentation of Probus V chapter 5.
    All three-digit errorcodes are from this python module.
    """

    E0 = "no error", "standard response on each command"
    E1 = (
        "no data available",
        (
            "Customer tried to read from GPIB but there were no data prepared. "
            "(IBIG50 sent command ~T2 to ADDA)"
        ),
    )
    E2 = "unknown register type", "No valid register type after '>'"
    E4 = (
        "invalid argument",
        "The argument of the command was rejected .i.e. malformed number",
    )
    E5 = "argument out of range", "i.e. setvalue higher than type value"
    E6 = (
        "register is read only",
        "Some registers can only be read but not written to. (i.e. monitor registers)",
    )
    E7 = "Receive Overflow", "Command string was longer than 50 characters."
    E8 = (
        "EEPROM is write protected",
        (
            "Write attempt to calibration data while the write protection switch was"
            " set to write protected."
        ),
    )
    E9 = (
        "address error",
        (
            "A non addressed command was sent to ADDA while it was in addressable mode "
            "(and vice versa)."
        ),
    )
    E10 = "unknown SCPI command", "This SCPI command is not implemented"
    E11 = (
        "not allowed Trigger-on-Talk",
        (
            "Not allowed attempt to Trigger-on-Talk (~T1) while ADDA was in addressable"
            " mode."
        ),
    )
    E12 = "invalid argument in ~Tn command", "Only ~T1 and ~T2 is implemented."
    E13 = (
        "invalid N-value",
        (
            "Register > K8 contained an invalid value. Error code is output on an"
            " attempt to query data with ? or ~T1"
        ),
    )
    E14 = "register is write only", "Some registers can only be writte to (i.e.> H0)"
    E15 = "string too long", "i.e.serial number string too long during calibration"
    E16 = (
        "wrong checksum",
        (
            "checksum over command string was not correct, refer also to 4.4 of the "
            "Probus V documentation"
        ),
    )
    E100 = (
        "Command is not implemented",
        "You tried to execute a command, which is not implemented or does not exist",
    )
    E106 = (
        "The rampstate is a read-only register",
        (
            "You tried to write data to the register, which can only give "
            "you the status of the ramping."
        ),
    )
    E206 = (
        "This status register is read-only",
        (
            "You tried to write data to this "
            "register, which can only give you "
            "the actual status of the "
            "corresponding digital output."
        ),
    )
    E306 = (
        "The monitor register is read-only",
        "You tried to write data to a monitor, which can only give you measured data.",
    )
    E115 = (
        "The given index to select a digital value is out of range",
        "Only integer values between 0 and 1 are allowed.",
    )
    E125 = (
        "The given index to select a ramp mode is out of range",
        "Only integer values between 0 and 4 are allowed.",
    )
    E135 = (
        "The given index to select the readback channel is out of range",
        "Only integer values between 0 and 6 are allowed.",
    )
    E145 = (
        "The given value for the AD-conversion is unknown",
        'Valid values for the ad-conversion are integer values from "0" to "7".',
    )
    E155 = (
        "The given value to select a polarity is out range.",
        "The value should be 0 or 1.",
    )
    E165 = "The given index to select the terminator string is out of range", ""
    E504 = "Empty string as response", "The connection is broken."
    E505 = (
        "The returned register is not the requested.",
        "Maybe the connection is overburden.",
    )
    E666 = (
        (
            "You cannot overwrite the most recent error in the interface of the power"
            " supply. But, well: You created an error anyway..."
        ),
        "",
    )

    def raise_(self) -> None:
        """
        Evaluate the error code and raise a `FuGError` if the error code is not `E0`
        """
        if self is FuGErrorcodes.E0:
            logger.debug('Communication with FuG successful, errorcode "E0" received.')
            return
        logger.debug(f"A FuGError with the errorcode {self.name} was detected.")
        msg = f"{self.description}. Possible reason: {self.possible_reason}"
        logger.exception(msg)
        raise FuGError(msg, errorcode=self.name)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Mock OscilloscopeTrigger

"""

from libtiepie.oscilloscope import Oscilloscope as LtpOscilloscopeTrigger


class OscilloscopeTrigger(LtpOscilloscopeTrigger):
    """"""

    def __init__(self) -> None:
        self._timeout = None

    def __del__(self) -> None:
        pass

    def _get_timeout(self):
        """ """
        return self._timeout

    def _set_timeout(self, value):
        self._timeout = value

    def verify_timeout(self, value):
        """Verify if a required trigger time out can be set, without actually setting
        the hardware itself.

        :param value: The required trigger time out in seconds,
            or #TIEPIE_HW_TO_INFINITY.
        :returns: The trigger time out that would have been set,
            if tiepie_hw_oscilloscope_trigger_set_timeout() was used.
        .. version added:: 1.0
        """
        return value

    def verify_timeout_ex(self, value, measure_mode, sample_rate) -> None:
        """Verify if a required trigger time out can be set, without actually setting
        the hardware itself.

        :param value: The required trigger time out in seconds, or
            #TIEPIE_HW_TO_INFINITY.
        :param measure_mode: Measure mode, a TIEPIE_HW_MM_* value.
        :param sample_rate: Sample rate in Hz.
        :returns: The trigger time out that would have been set,
            if tiepie_hw_oscilloscope_trigger_set_timeout() was used.
        .. version added:: 1.0
        """

    def _get_has_delay(self):
        """ """
        return

    def has_delay_ex(self, measure_mode) -> None:
        """Check whether the oscilloscope has trigger delay support for a specified
        measure mode.

        :param measure_mode: Measure mode, a TIEPIE_HW_MM_* value.
        :returns: ``True`` if the oscilloscope has trigger delay support,
            ``False`` otherwise.
        .. version added:: 1.0
        """

    def _get_delay_max(self):
        """Maximum trigger delay in seconds, for the currently selected measure mode and
        sample rate."""
        return

    def get_delay_max_ex(self, measure_mode, sample_rate) -> None:
        """Get the maximum trigger delay in seconds, for a specified measure mode and
        sample rate.

        :param measure_mode: Measure mode, a TIEPIE_HW_MM_* value.
        :param sample_rate: Sample rate in Hz.
        :returns: The maximum trigger delay in seconds.
        .. version added:: 1.0
        """

    def _get_delay(self):
        """Currently selected trigger delay in seconds."""
        return

    def _set_delay(self, value): ...

    def verify_delay(self, value) -> None:
        """Verify if a required trigger delay can be set, without actually setting the
        hardware itself.

        :param value: The required trigger delay in seconds.
        :returns: The trigger delay that would have been set,
            if tiepie_hw_oscilloscope_trigger_set_delay() was used.
        .. version added:: 1.0
        """

    def verify_delay_ex(self, value, measure_mode, sample_rate) -> None:
        """Verify if a required trigger delay can be set, without actually setting the
        hardware itself.

        :param value: The required trigger delay in seconds.
        :param measure_mode: Measure mode, a TIEPIE_HW_MM_* value.
        :param sample_rate: Sample rate in Hz.
        :returns: The trigger delay that would have been set,
            if tiepie_hw_oscilloscope_trigger_set_delay() was used.
        .. version added:: 1.0
        """

    timeout = property(_get_timeout, _set_timeout)
    has_delay = property(_get_has_delay)
    delay_max = property(_get_delay_max)
    delay = property(_get_delay, _set_delay)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Module for controlling pulse output channels A, B, C and D.
"""

import logging
import re

from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_bool, validate_number

from .base import Polarity, T560Error, _ChannelStatus

logger = logging.getLogger(__name__)


class _Channel:
    """
    A T560 channel
    """

    def __init__(self, device, name) -> None:
        self.ch_name = name
        self.device = device  # The T560 object

    @property
    def _status(self) -> _ChannelStatus:
        """
        Get the settings of the channel from the T560. Example response:
        "Ch A  POS  ON  Dly  00.000,000,000,000  Wid  00.000,002,000,000"
        """
        response = self.device.com.query(f"{self.ch_name}S")
        # Check for valid response with regex
        pattern_name = rf"Ch\s*{self.ch_name}"
        pattern_polarity = r"(?P<polarity>\w+)"
        pattern_enabled = r"(?P<enabled>\w+)"
        pattern_delay = r"Dly\s*(?P<delay>[0-9,.]+)"
        pattern_width = r"Wid\s*(?P<width>[0-9,.]+)"
        pattern = (
            rf"{pattern_name}\s*{pattern_polarity}\s*{pattern_enabled}\s*"
            rf"{pattern_delay}\s*{pattern_width}"
        )
        result = re.search(pattern, response)
        if not result:
            msg = f"Cannot identify the channel status. Got: '{response}'"
            logger.error(msg)
            raise T560Error(msg)

        polarity = Polarity(result.group("polarity"))
        enabled = result.group("enabled") == "ON"
        delay = float(result.group("delay").replace(",", "_"))
        width = float(result.group("width").replace(",", "_"))
        return _ChannelStatus(
            polarity=polarity, enabled=enabled, delay=delay, width=width
        )

    @property
    def enabled(self) -> bool:
        """
        Channel output ON/OFF
        """
        enabled = self._status.enabled
        if enabled:
            logger.info(f"Ch {self.ch_name} is enabled")
        else:
            logger.info(f"Ch {self.ch_name} is disabled")
        return enabled

    @enabled.setter
    def enabled(self, enabled: bool):
        """
        Enable or disable the channel.

        :param enabled: channel ON/OFF status as a bool
        :raises TypeError: if enabled is not a bool-like
        """
        validate_bool("enabled", enabled, logger=logger)
        if enabled:
            self.device.com.query(f"{self.ch_name}S ON")
            logger.info(f"Ch {self.ch_name} has been enabled")
        else:
            self.device.com.query(f"{self.ch_name}S OFF")
            logger.info(f"Ch {self.ch_name} has been disabled")

    @property
    def polarity(self) -> Polarity:
        """
        "POS"": the channel is active-high
        "NEG": the channel is active-low
        """
        polarity = self._status.polarity
        logger.info(f"Ch {self.ch_name} polarity: {polarity}")
        return polarity

    @polarity.setter
    def polarity(self, polarity: str | Polarity):
        """
        Set the polarity of the channel.

        :param polarity: "POS" or "NEG", see Polarity enum
        :raises ValueError: if polarity is not in Polarity enum
        """
        self.device.com.query(f"{self.ch_name}S {Polarity(polarity)}")
        logger.info(f"Ch {self.ch_name} polarity set to {polarity}")

    @property
    def delay(self) -> Number:
        """
        The time between the trigger and the leading edge of the pulse.
        """
        delay = self._status.delay
        logger.info(f"Ch {self.ch_name} pulse delay: {delay} s")
        return delay

    @delay.setter
    def delay(self, delay: Number):
        """
        Set the time between the trigger and the leading edge of the pulse.

        :param delay: pulse delay in seconds, 10 ps resolution.
        :raises ValueError: if delay is negative or greater than 10s
        """
        validate_number("delay", delay, limits=(0, 10), logger=logger)
        self.device.com.query(f"{self.ch_name}D {delay:.12f}s")
        logger.info(f"Ch {self.ch_name} pulse delay set to {delay} s")

    @property
    def width(self) -> Number:
        """
        Duration of a pulse.
        """
        width = self._status.width
        logger.info(f"Ch {self.ch_name} pulse width: {width} s")
        return width

    @width.setter
    def width(self, width: Number):
        """
        Set the duration of a pulse.

        :param width: pulse width in seconds, 10 ps resolution.
        :raises ValueError: if width is negative or greater than 10s
        """
        validate_number("width", width, limits=(0, 10), logger=logger)
        self.device.com.query(f"{self.ch_name}W {width:.12f}s")
        logger.info(f"Ch {self.ch_name} pulse width set to {width} s")



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the .dev.mbw973 sub-package.
"""

import pytest

from hvl_ccb.comm.serial import SerialCommunicationIOError
from hvl_ccb.dev.mbw973 import (
    MBW973,
    MBW973Config,
    MBW973ControlRunningError,
    MBW973PumpRunningError,
    MBW973SerialCommunicationConfig,
)
from masked_comm.serial import LoopSerialCommunication


@pytest.fixture(scope="module")
def com_config():
    return {
        "port": "loop://?logging=debug",
        "baudrate": 115200,
        "parity": MBW973SerialCommunicationConfig.Parity.NONE,
        "stopbits": MBW973SerialCommunicationConfig.Stopbits.ONE,
        "bytesize": MBW973SerialCommunicationConfig.Bytesize.EIGHTBITS,
        "terminator": b"\r\n",
        "timeout": 3,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {"polling_interval": 0.1}


@pytest.fixture
def started_mbw973_device(com_config, dev_config):
    serial_port = LoopSerialCommunication(com_config)
    serial_port.open()
    serial_port.put_text("1")
    serial_port.put_text("0")
    with MBW973(serial_port, dev_config) as mbw:
        while serial_port.get_written() is not None:
            pass
        yield serial_port, mbw


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    MBW973Config()

    config = MBW973Config(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"polling_interval": 0},
        {"polling_interval": -1},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        MBW973Config(**invalid_config)


def test_instantiation(com_config, dev_config) -> None:
    mbw = MBW973(com_config)
    assert mbw is not None

    mbw = MBW973(com_config, dev_config)
    assert mbw is not None


def test_com_error(com_config, dev_config) -> None:
    wrong_config = dict(com_config)
    wrong_config["port"] = "NOT A PORT"
    mbw = MBW973(wrong_config, dev_config)
    assert not mbw.com.is_open

    with pytest.raises(SerialCommunicationIOError):
        mbw.start()

    mbw = MBW973(com_config, dev_config)
    assert not mbw.com.is_open

    with pytest.raises(SerialCommunicationIOError):
        mbw.write(0)
    with pytest.raises(SerialCommunicationIOError):
        mbw.is_done()
    with pytest.raises(SerialCommunicationIOError):
        mbw.start_control()
    with pytest.raises(SerialCommunicationIOError):
        mbw.read_measurements()
    with pytest.raises(SerialCommunicationIOError):
        mbw.set_measuring_options()


def test_start(started_mbw973_device) -> None:
    com, mbw = started_mbw973_device

    # starting again should work
    com.put_text("1")
    com.put_text("0")
    mbw.start()
    assert com.get_written() == "HumidityTest?"
    assert com.get_written() == "SF6PurityTest?"


def test_is_done(started_mbw973_device) -> None:
    serial_port, mbw = started_mbw973_device

    # case 1, only humidity
    mbw.set_measuring_options(humidity=True, sf6_purity=False)
    assert serial_port.get_written() == "HumidityTest=1"
    assert serial_port.get_written() == "SF6PurityTest=0"
    serial_port.put_text("0")
    assert mbw.is_done() is False
    assert serial_port.get_written() == "DoneWithDP?"

    # case 2, only SF6 purity
    mbw.set_measuring_options(humidity=False, sf6_purity=True)
    assert serial_port.get_written() == "HumidityTest=0"
    assert serial_port.get_written() == "SF6PurityTest=1"
    serial_port.put_text("0")
    assert mbw.is_done() is False
    assert serial_port.get_written() == "SF6VolHold?"

    # case 3, both measurements
    mbw.set_measuring_options(humidity=True, sf6_purity=True)
    assert serial_port.get_written() == "HumidityTest=1"
    assert serial_port.get_written() == "SF6PurityTest=1"
    serial_port.put_text("1")
    serial_port.put_text("0")
    assert mbw.is_done() is False
    assert serial_port.get_written() == "DoneWithDP?"
    assert serial_port.get_written() == "SF6VolHold?"

    # case 3, both measurements, resulting in True
    mbw.set_measuring_options(humidity=True, sf6_purity=True)
    assert serial_port.get_written() == "HumidityTest=1"
    assert serial_port.get_written() == "SF6PurityTest=1"
    serial_port.put_text("1")
    serial_port.put_text("1")
    serial_port.put_text("-17.2315")
    serial_port.put_text("-21.5501")
    serial_port.put_text("170.421")
    serial_port.put_text("87.2323")
    serial_port.put_text("35.0001")
    serial_port.put_text("87.53")
    assert mbw.is_done() is True
    assert mbw.last_measurement_values == {
        "frostpoint": -17.2315,
        "frostpoint_ambient": -21.5501,
        "pressure": 170.421,
        "ppmv": 87.2323,
        "ppmw": 35.0001,
        "sf6_vol": 87.53,
    }

    mbw.stop()


def test_mbw973_start_control(started_mbw973_device) -> None:
    """
    Tests for the start_control method
    """

    serial_port, mbw = started_mbw973_device

    # control still running
    serial_port._serial_port.reset_input_buffer()
    serial_port.put_text("1")
    with pytest.raises(MBW973ControlRunningError):
        mbw.start_control()

    assert serial_port.get_written() == "control?"

    # pump still running
    serial_port.put_text("0")
    serial_port.put_text("1")
    with pytest.raises(MBW973PumpRunningError):
        mbw.start_control()

    assert serial_port.get_written() == "control?"
    assert serial_port.get_written() == "Pump.on?"

    # nothing running, can start
    serial_port.put_text("0")  # no control running
    serial_port.put_text("0")  # no pump running
    mbw.start_control()
    assert serial_port.get_written() == "control?"
    assert serial_port.get_written() == "Pump.on?"
    assert serial_port.get_written() == "control=1"

    serial_port.put_text("0")  # not done with dewpoint
    assert mbw.is_done_with_measurements is False
    serial_port.put_text("1")  # done with dewpoint
    # dummy measurement values in the readout order
    # note: for Py >= 3.6 dict items are ordered by default (no need to use OrderedDict)
    expected_measurement_values = {
        "frostpoint": 0.1,
        "frostpoint_ambient": 0.2,
        "pressure": 0.3,
        "ppmv": 0.4,
        "ppmw": 0.5,
        "sf6_vol": 0.6,
    }
    for measurement_value in expected_measurement_values.values():
        serial_port.put_text(str(measurement_value))
    mbw.status_poller.wait_for_polling_result()
    assert mbw.is_done_with_measurements is True
    assert mbw.last_measurement_values == expected_measurement_values


def test_read_measurements(started_mbw973_device) -> None:
    com, mbw = started_mbw973_device

    expected_measurement_values = {
        "frostpoint": 0.1,
        "frostpoint_ambient": 0.2,
        "pressure": 0.3,
        "ppmv": 0.4,
        "ppmw": 0.5,
        "sf6_vol": 0.6,
    }
    for measurement_value in expected_measurement_values.values():
        com.put_text(str(measurement_value))

    readout = mbw.read_measurements()

    assert readout == expected_measurement_values

    assert com.get_written() == "Fp?"
    assert com.get_written() == "Fp1?"
    assert com.get_written() == "Px?"
    assert com.get_written() == "PPMv?"
    assert com.get_written() == "PPMw?"
    assert com.get_written() == "SF6Vol?"


def test_set_measuring_options(started_mbw973_device) -> None:
    test_com, mbw = started_mbw973_device

    mbw.set_measuring_options(True, True)
    assert test_com.get_written() == "HumidityTest=1"
    assert test_com.get_written() == "SF6PurityTest=1"

    mbw.set_measuring_options(True, False)
    assert test_com.get_written() == "HumidityTest=1"
    assert test_com.get_written() == "SF6PurityTest=0"

    mbw.set_measuring_options(False, True)
    assert test_com.get_written() == "HumidityTest=0"
    assert test_com.get_written() == "SF6PurityTest=1"

    mbw.set_measuring_options(False, False)
    assert test_com.get_written() == "HumidityTest=0"
    assert test_com.get_written() == "SF6PurityTest=0"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
EarthingStick of the different "Cubes".
"""

import logging

from .constants import SafetyStatus
from .errors import CubeEarthingStickOperationError
from .switches import (  # noqa: F401
    SwitchOperatingStatus,
    SwitchOperation,
    SwitchStatus,
    _Switch,
)

logger = logging.getLogger(__name__)


class _EarthingStick(_Switch):
    """
    Earthing sticks with status, operating status (manual and auto) and manual operate.
    """

    _STICKS: tuple = (1, 2, 3, 4, 5, 6)
    _ERROR_CLS = CubeEarthingStickOperationError
    _SWITCHABLE_AT_STATES = (  # type: ignore[assignment]
        SafetyStatus.RED_READY,
        SafetyStatus.RED_OPERATE,
    )

    def __init__(self, handle, number: int) -> None:
        super().__init__(handle, device_name=f"Earthing Stick {number}")

        self._CMD_STATUS: str = (
            f'"DB_Safety_Circuit"."Earthstick_{number}"."si_HMI_Status"'
        )
        self._CMD_OPERATING_STATUS: str = (
            f'"DB_Safety_Circuit"."Earthstick_{number}"."sx_manual_control_active"'
        )
        self._CMD_MANUAL: str = (
            f'"DB_Safety_Circuit"."Earthstick_{number}"."sx_earthing_manually"'
        )



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for PT104
"""

import pytest

from hvl_ccb.dev.picotech_pt104 import Pt104, Pt104CommunicationType, Pt104DeviceConfig


@pytest.fixture(scope="module")
def com_config():
    return {}


@pytest.fixture(scope="module")
def dev_config():
    return {
        "host": "127.0.0.1",
        "port": 6249,
        "serial_number": "HS337/135",
        "interface": Pt104CommunicationType.ETHERNET,
    }


def test_instantiation(dev_config) -> None:
    dev = Pt104({}, dev_config)
    assert dev is not None


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {
            "interface": Pt104CommunicationType.ETHERNET,
            "host": None,
        },
    ],
)
def test_dev_config_invalid(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(AttributeError):
        Pt104DeviceConfig(**invalid_config)


def test_ct_requires_host() -> None:
    assert Pt104CommunicationType.ETHERNET.requires_host
    assert not Pt104CommunicationType.USB.requires_host



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#

import enum
import logging
from types import FunctionType
from typing import Any, cast

import aenum
import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from hvl_ccb.utils.validation import validate_number

logger = logging.getLogger(__name__)


def preserve_type(func):  # noqa: ANN201
    """
    This wrapper preserves the first order type of the input.
    Upto now the type of the data stored in a list, tuple, array or dict
    is not preserved.
    Integer will be converted to float!
    """

    def wrap(self, value, **kwargs):
        validate_number("value", value, (-np.inf, np.inf), logger=logger)

        is_dict = False

        if isinstance(value, dict):
            keys = list(value.keys())
            value = list(value.values())
            is_dict = True

        value_func = np.asarray(value, dtype=float)

        value_func = func(self, value_func, **kwargs)

        if is_dict:  # as 'value' is now list
            value = dict(zip(keys, list(value_func), strict=False))
        elif isinstance(value, list):
            value = list(value_func)
        elif isinstance(value, tuple) and hasattr(value, "_fields"):
            value = type(value)(*value_func)
        elif isinstance(value, tuple):
            value = tuple(value_func)
        elif isinstance(value, int | float) and not isinstance(value, np.generic):
            value = float(value_func)
        else:
            value = value_func
        return value

    return wrap


def convert_value_to_str(*value: npt.NDArray) -> list[str | list[str]]:
    """
    Converts two sets of values to strings. This is necessary because a 0-dim array
    needs different treatment than a 1-dim array

    :param value: array of values either 0-dim or 1-dim
    :return:converted se
    """
    value_str: list[str | list[str]] = []
    for val in value:
        if val.ndim == 0:
            value_str.append(f"{val:.2f}")
        else:
            value_str.append([f"{v:.2f}" for v in val])
    return cast("list[str | list[str]]", value_str)


class GetAttr:
    def __init__(self, default, name) -> None:
        self._value = default
        self._name = name

    def __get__(self, instance, owner) -> Self | Any:
        if not instance:
            return self
        return self._value

    def __set__(self, instance, value) -> None:
        msg = (
            f"It is not possible to set the value of '{self._name}'. "
            "This parameter is readonly!"
        )
        logger.error(msg)
        raise AttributeError(msg)


class SetAttr(GetAttr):
    def __init__(
        self, default, name, limits, absolut=False, dtype=(int, float), validator=None
    ) -> None:
        super().__init__(default, name)
        self._limits = limits
        self._absolut = absolut
        self._dtype = dtype
        self._validator = validator

    def __set__(self, instance, value) -> None:
        if (
            type(self._validator) is FunctionType
            and self._validator.__name__ == "validate_number"
        ):
            if self._absolut:
                self._validator(self._name, abs(value), self._limits, self._dtype)
            else:
                self._validator(self._name, value, self._limits, self._dtype)
        elif issubclass(self._validator, enum.Enum | aenum.Enum):
            self._validator(value)
        else:
            msg = "Unknown validator, please implement this version"
            logger.error(msg)
            raise NotImplementedError(msg)
        self._value = value



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Controller for the Pico Technology PT-104 temperature logger device.
The controller is written as a wrapper around Pico Technology driver for the PT-104
device.

This code is directly based on: https://github.com/trombastic/Pico_PT104/ .

Extra installation
~~~~~~~~~~~~~~~~~~

Pico Technology driver for the PT-104 device is available only on Windows and on Linux.

To use this PT-104 device wrapper:

1. install the :code:`hvl_ccb` package with a :code:`picotech` extra feature::

        $ pip install "hvl_ccb[picotech]"

   this will install the Python bindings for the library.

2. install the library

    * on Windows: download and install PicoSDK from https://www.picotech.com/downloads
      (choose "PicoLog Data Loggers" > "PT-104" > "Software");
    * on Linux:
        - for Ubuntu/Debian, install :code:`libusbpt104` from :code:`.deb` file found in
          https://labs.picotech.com/debian/pool/main/libu/libusbpt104/ (note: at the
          moment the PT-104 driver is not a part of the official :code:`picoscope`
          package; cf.
          https://www.picotech.com/support/topic40626.html );
        - for any other supported Linux distribution, follow instructions to install
          the "USB PT-104 devices" drivers in https://www.picotech.com/downloads/linux ;

"""

import logging
from abc import ABC
from ctypes import byref, c_long, c_short, c_ushort
from dataclasses import dataclass, field
from enum import IntEnum
from functools import wraps
from ipaddress import IPv4Address, IPv6Address
from time import time

import aenum
from picosdk.errors import (  # type: ignore[import-untyped,import-not-found]
    CannotFindPicoSDKError,
)

try:
    from picosdk.usbPT104 import (  # type: ignore[import-untyped,import-not-found]
        usbPt104,
    )
except CannotFindPicoSDKError as exc:
    # PicoSDK Python wrapper tries to import PicoSDK system library already on import
    raise ImportError from exc

from hvl_ccb.comm import NullCommunicationProtocol
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError, SingleCommDevice
from hvl_ccb.utils.validation import validate_and_resolve_host, validate_tcp_port

logger = logging.getLogger(__name__)


class Pt104Error(DeviceError):
    """
    Error to indicate communication issues with the pt104.
    """


def require_started(method):  # noqa: ANN201
    """
    Check if device `is_started` and raise an `Pt104Error` if not.

    :param method: `Pt104` instance method to wrap
    :return: Whatever `method` returns
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.is_started:
            err_msg = f"{self} is not started. Run `start()` method first."
            logger.error(err_msg)
            raise Pt104Error(err_msg)
        return method(self, *args, **kwargs)

    return wrapper


class Pt104Channels(IntEnum):
    CHANNEL_1 = 1
    CHANNEL_2 = 2
    CHANNEL_3 = 3
    CHANNEL_4 = 4
    CHANNEL_5 = 5
    CHANNEL_6 = 6
    CHANNEL_7 = 7
    CHANNEL_8 = 8
    MAX_CHANNELS = CHANNEL_8


class Pt104Wires(IntEnum):
    WIRES_2 = 2
    WIRES_3 = 3
    WIRES_4 = 4
    MIN_WIRES = WIRES_2
    MAX_WIRES = WIRES_4


class Pt104DataTypes(IntEnum):
    OFF = 0
    PT100 = 1
    PT1000 = 2
    RESISTANCE_TO_375R = 3
    RESISTANCE_TO_10K = 4
    DIFFERENTIAL_TO_115MV = 5
    DIFFERENTIAL_TO_2500MV = 6
    SINGLE_ENDED_TO_115MV = 7
    SINGLE_ENDED_TO_2500MV = 8

    @property
    def is_active(self) -> bool:
        return self is not Pt104DataTypes.OFF


_pt104_ct_dict = usbPt104.COMMUNICATION_TYPE()


class Pt104CommunicationType(aenum.IntEnum, init="value open_unit"):  # type: ignore[call-arg]
    USB = (
        _pt104_ct_dict["CT_USB"],
        usbPt104._OpenUnit,  # UsbPt104OpenUnit
    )
    ETHERNET = (
        _pt104_ct_dict["CT_ETHERNET"],
        usbPt104._OpenUnitViaIp,  # UsbPt104OpenUnitViaIP
    )

    @property
    def requires_host(self) -> bool:
        return self == self.__class__.ETHERNET


@configdataclass
class Pt104DeviceConfig:
    """
    Configuration dataclass for PT104
    """

    #: Serial number
    serial_number: str
    #: Interface to communicate on
    interface: Pt104CommunicationType
    #: Host TCP/IP address, if applicable
    host: str | IPv4Address | IPv6Address | None = None
    #: TCP port number, if applicable
    port: int | None = None

    def clean_values(self) -> None:
        if self.interface.requires_host:
            self.force_value(  # type: ignore[attr-defined]
                "host",
                validate_and_resolve_host(self.host, logger),  # type: ignore[arg-type]
            )
            validate_tcp_port(self.port, logger)

    @property
    def host_address(self) -> str | None:
        if self.host:
            port_sufix = f":{self.port}" if self.port else ""
            return f"{self.host}{port_sufix}"
        return None


@dataclass
class Pt104ChannelConfig:
    data_type: Pt104DataTypes = Pt104DataTypes.OFF
    nb_wires: Pt104Wires = Pt104Wires.WIRES_4
    low_pass_filter: bool = True
    value: c_long = field(default_factory=lambda: c_long(0))
    last_query: float = field(init=False)

    def __post_init__(self) -> None:
        self.last_query = time()


class Pt104(SingleCommDevice, ABC):
    """
    PicoTech pt104 data logger class.
    """

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for pt104.

        :param com: object to use as communication protocol.
        """

        # Call superclass constructor
        super().__init__(com, dev_config)

        self.channels: dict[Pt104Channels, Pt104ChannelConfig] = {
            ch: Pt104ChannelConfig()
            for ch in (
                Pt104Channels.CHANNEL_1,
                Pt104Channels.CHANNEL_2,
                Pt104Channels.CHANNEL_3,
                Pt104Channels.CHANNEL_4,
            )
        }
        self._handle: c_short | None = None

    @staticmethod
    def default_com_cls() -> type[NullCommunicationProtocol]:
        return NullCommunicationProtocol

    @staticmethod
    def config_cls():
        return Pt104DeviceConfig

    def _validate_channel(self, channel: int) -> None:
        if channel not in self.channels:
            msg = (
                f"Invalid channel {channel}; available channels are: "
                f"{', '.join([str(ch) for ch in self.channels])}."
            )
            raise ValueError(msg)

    @property
    def is_connected(self) -> bool:
        """Check the connection status.

        :return: `True` if connection with device is initiated, `False` otherwise.
        """
        return self._handle is not None

    @property
    def is_started(self) -> bool:
        """Check if device is started.

        :return: `True` if device is started, `False` otherwise.
        """
        return self.is_connected

    def start(self) -> None:
        """Connect to a Pt-104A data acquisition module via USB or Ethernet"""
        logger.info(f"Starting device {self}")
        super().start()

        if self.is_connected:
            self.stop()

        self._handle = c_short()

        open_unit_extra_args = []
        if self.config.interface.requires_host:
            open_unit_extra_args.append(self.config.host_address.encode())
        status_unit = self.config.interface.open_unit(
            byref(self._handle),
            self.config.serial_number.encode(),
            *open_unit_extra_args,
        )

        if status_unit != 0:
            self._handle = None
            msg = (
                "Communication with pt104 could not be established, "
                f"error code: {status_unit}"
            )
            logger.error(msg)
            raise Pt104Error(msg)

        self.set_channels()

    @property
    def active_channel_count(self) -> int:
        """return the number of active channels
        :return: number of active channels
        """
        n = 0
        for conf in self.channels.values():
            if conf.data_type.is_active:
                n += 1
        return n

    def stop(self) -> None:
        """disconnect from the unit
        :return: bool
        """
        logger.info(f"Stopping device {self}")
        usbPt104.UsbPt104CloseUnit(self._handle)
        self._handle = None
        super().stop()

    @require_started
    def set_channel(
        self,
        channel: Pt104Channels,
        data_type: Pt104DataTypes,
        nb_wires: Pt104Wires,
        low_pass_filter: bool = True,
    ) -> None:
        """writes the channel configuration to self.channels and the device.
        :param channel: channel number (Pt104Channels)
        :param data_type: data type of the connected probe (DataType)
        :param nb_wires: number of wires (Pt104Wires)
        :param low_pass_filter: use the low pass filter [True, False]
        :return: status
        """
        logger.info(f"Configuring channel: {channel}")
        self._validate_channel(channel)
        self.channels[channel].data_type = data_type
        self.channels[channel].nb_wires = nb_wires
        self.channels[channel].low_pass_filter = low_pass_filter

        cs = usbPt104.UsbPt104SetChannel(self._handle, channel, data_type, nb_wires)
        if cs != 0:
            msg = f"Setting channel {channel} failed with error {cs}"
            logger.error(msg)
            raise Pt104Error(msg)

    def set_channels(self) -> None:
        """sets the channel configuration from self.channels"""
        for channel in self.channels:
            self.set_channel(
                channel,
                self.channels[channel].data_type,
                self.channels[channel].nb_wires,
            )

    @require_started
    def get_value(self, channel: Pt104Channels, raw_value: bool = False) -> float:
        """queries the measurement value from the unit
        :param channel: channel number (Pt104Channels)
        :param raw_value: skip conversion
        :return: measured value
        """
        self._validate_channel(channel)
        result = None
        while result is None:
            self._wait_for_conversion(channel)
            status_channel = usbPt104.UsbPt104GetValue(
                self._handle,
                channel,
                byref(self.channels[channel].value),
                self.channels[channel].low_pass_filter,
            )
            self.channels[channel].last_query = time()
            if status_channel == 0:
                if raw_value:
                    return float(self.channels[channel].value.value)
                return self.scale_value(
                    float(self.channels[channel].value.value), channel
                )

            msg = f"Error while reading {channel}, error code {status_channel}"
            logger.error(msg)
            raise Pt104Error(msg)
        return None

    @property
    def get_value_channel_1(self) -> float:
        """queries the measurement value from channel 1
        :return: scaled measured value
        """
        return self.get_value(Pt104Channels.CHANNEL_1)

    @property
    def get_value_channel_2(self) -> float:
        """queries the measurement value from channel 2
        :return: scaled measured value
        """
        return self.get_value(Pt104Channels.CHANNEL_2)

    @property
    def get_value_channel_3(self) -> float:
        """queries the measurement value from channel 3
        :return: scaled measured value
        """
        return self.get_value(Pt104Channels.CHANNEL_3)

    @property
    def get_value_channel_4(self) -> float:
        """queries the measurement value from channel 4
        :return: scaled measured value
        """
        return self.get_value(Pt104Channels.CHANNEL_4)

    def set_mains(self, sixty_hertz: bool = False) -> None:
        """This function is used to inform the driver of the
        local mains (line) frequency.
        This helps the driver to filter out electrical noise.
        :param sixty_hertz: mains frequency is sixty
        :return: success
        """
        if sixty_hertz:
            sixty = c_ushort(1)
            logger.info("set mains freq to 60 Hz")
        else:
            sixty = c_ushort(0)
            logger.info("set mains freq to 50 Hz")
        usbPt104.UsbPt104SetMains(self._handle, sixty)

    def _wait_for_conversion(self, channel: Pt104Channels) -> None:
        """wait until the adc conversion is finished
        :param channel: channel number (Pt104Channels)
        :return:
        """
        conversion_time = self.active_channel_count * 0.75
        last_query = self.channels[channel].last_query
        while last_query + conversion_time > time():
            pass

    def scale_value(self, value: float, channel: Pt104Channels) -> float:
        """scales the value from the device.
        :param value: value to convert as float
        :param channel: channel number (Pt104Channels)
        :return: Temperature in °C, Resistance in mOhm, Voltage in mV
        """
        if self.channels[channel].data_type in [
            Pt104DataTypes.PT100,
            Pt104DataTypes.PT1000,
        ]:
            return value / 10.0**3  # °C
        if self.channels[channel].data_type == Pt104DataTypes.RESISTANCE_TO_375R:
            return value / 10.0**3  # mOhm
        if self.channels[channel].data_type == Pt104DataTypes.RESISTANCE_TO_10K:
            return value  # mOhm
        if self.channels[channel].data_type in [
            Pt104DataTypes.DIFFERENTIAL_TO_115MV,
            Pt104DataTypes.SINGLE_ENDED_TO_115MV,
        ]:
            return value / 10.0**9  # mV
        if self.channels[channel].data_type in [
            Pt104DataTypes.DIFFERENTIAL_TO_2500MV,
            Pt104DataTypes.SINGLE_ENDED_TO_2500MV,
        ]:
            return value / 10.0**8  # mV

        msg = "selected datatype not recognized"
        raise Pt104Error(msg)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Tests for the Elektro Automatik PSI 9000 power supply device classes.
"""

import pytest

from hvl_ccb.dev.ea_psi9000 import (
    PSI9000,
    PSI9000Config,
    PSI9000Error,
    PSI9000VisaCommunicationConfig,
)
from masked_comm.visa import MaskedVisaCommunication


@pytest.fixture(scope="module")
def com_config():
    return {
        "interface_type": PSI9000VisaCommunicationConfig.InterfaceType.TCPIP_SOCKET,
        "host": "127.0.0.1",
        "open_timeout": 10,
        "timeout": 50,
    }


@pytest.fixture(scope="module")
def dev_config():
    return {
        "spoll_interval": 0.01,
        "spoll_start_delay": 0,
        "wait_sec_system_lock": 0.01,
        "wait_sec_settings_effect": 0.01,
        "wait_sec_initialisation": 0.01,
    }


@pytest.fixture
def testdev(com_config, dev_config):
    com = MaskedVisaCommunication(com_config)
    return PSI9000(com, dev_config)


def test_dev_config(dev_config) -> None:
    # currently there are no non-default config values
    PSI9000Config()

    config = PSI9000Config(**dev_config)
    for key, value in dev_config.items():
        assert getattr(config, key) == value


@pytest.mark.parametrize(
    "wrong_config_dict",
    [
        {"power_limit": -1.0},
        {"voltage_lower_limit": 20, "voltage_upper_limit": 19.0},
        {
            "voltage_lower_limit": 4,
            "voltage_upper_limit": PSI9000Config._VOLTAGE_UPPER_LIMIT + 1,
        },
        {"current_lower_limit": -1.0},
        {
            "current_lower_limit": 4,
            "current_upper_limit": PSI9000Config._CURRENT_UPPER_LIMIT + 1,
        },
        {"wait_sec_system_lock": 0},
        {"wait_sec_system_lock": -1},
        {"wait_sec_settings_effect": 0},
        {"wait_sec_initialisation": 0},
    ],
)
def test_invalid_config_dict(dev_config, wrong_config_dict) -> None:
    invalid_config = dict(dev_config)
    invalid_config.update(wrong_config_dict)
    with pytest.raises(ValueError):
        PSI9000Config(**invalid_config)


def test_instantiation(com_config) -> None:
    power_supply = PSI9000(com_config)
    assert power_supply is not None


def test_start_stop(testdev: PSI9000) -> None:
    power_supply = testdev
    test_com = power_supply.com

    power_supply.start()

    # stop successful
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    test_com.put_name("OUTput?", "OFF")
    power_supply.stop()

    # stop successful: output and lock on
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    test_com.put_name("OUTPut?", "ON")
    test_com.put_name("SOURce:VOLTage?", "0 V")
    test_com.put_name("SOURce:CURRent?", "0 A")
    test_com.put_name("OUTPut?", "OFF")
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    power_supply.stop()


def test_locking(testdev: PSI9000) -> None:
    power_supply = testdev
    test_com = power_supply.com

    # lock failed
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(True)
    assert test_com.get_written() == "SYSTem:LOCK ON"

    # lock successful
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    power_supply.set_system_lock(True)
    assert test_com.get_written() == "SYSTem:LOCK ON"

    # unlock failed due to still on current
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    test_com.put_name("MEASure:VOLTage?", "20 V")
    test_com.put_name("MEASure:CURRent?", "100 A")
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(False)
    assert test_com.get_written() is None

    # unlock failed
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    test_com.put_name("MEASure:VOLTage?", "0 V")
    test_com.put_name("MEASure:CURRent?", "0 A")
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    with pytest.raises(PSI9000Error):
        power_supply.set_system_lock(False)
    assert test_com.get_written() == "SYSTem:LOCK OFF"

    # unlock successful
    test_com.put_name("SYSTem:LOCK:OWNer?", "REMOTE")
    test_com.put_name("MEASure:VOLTage?", "0 V")
    test_com.put_name("MEASure:CURRent?", "0 A")
    test_com.put_name("SYSTem:LOCK:OWNer?", "NONE")
    power_supply.set_system_lock(False)
    assert test_com.get_written() == "SYSTem:LOCK OFF"

    # illegal answer
    test_com.put_name("SYSTem:LOCK:OWNer?", "bla")
    with pytest.raises(PSI9000Error):
        power_supply.get_system_lock()


def test_output(testdev: PSI9000) -> None:
    power_supply = testdev
    test_com = power_supply.com

    test_com.put_name("OUTPut?", "ON")
    assert power_supply.get_output()

    test_com.put_name("OUTPut?", "OFF")
    assert not power_supply.get_output()

    test_com.put_name("OUTPut?", "bla")
    with pytest.raises(PSI9000Error):
        power_supply.get_output()

    # setting output
    test_com.put_name("OUTPut?", "ON")
    power_supply.set_output(True)
    assert test_com.get_written() == "OUTPut ON"

    test_com.put_name("OUTPut?", "OFF")
    power_supply.set_output(False)
    assert test_com.get_written() == "OUTPut OFF"

    test_com.put_name("OUTPut?", "ON")
    with pytest.raises(PSI9000Error):
        power_supply.set_output(False)
    assert test_com.get_written() == "OUTPut OFF"

    test_com.put_name("OUTPut?", "OFF")
    with pytest.raises(PSI9000Error):
        power_supply.set_output(True)
    assert test_com.get_written() == "OUTPut ON"


def test_get_uip_upper_limits(testdev: PSI9000) -> None:
    test_com = testdev.com

    # get upper limits
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "80 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "1000 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    assert testdev.get_uip_upper_limits() == (80, 1000, 10000)


def test_get_ui_lower_limits(testdev: PSI9000) -> None:
    test_com = testdev.com

    # get lower limits
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "2 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5 A")
    assert testdev.get_ui_lower_limits() == (2, 5)


def test_set_upper_limits(testdev: PSI9000) -> None:
    test_com = testdev.com

    # set upper limits, success
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "2 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5 A")
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "10 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "20 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() == "SOURce:VOLTage:LIMit:HIGH 10"
    assert test_com.get_written() == "SOURce:CURRent:LIMit:HIGH 20"
    assert test_com.get_written() == "SOURce:POWer:LIMit:HIGH 10000"

    # set upper limits, set fail
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "2 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5 A")
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "10 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "200 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    with pytest.raises(PSI9000Error):
        testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() == "SOURce:VOLTage:LIMit:HIGH 10"
    assert test_com.get_written() == "SOURce:CURRent:LIMit:HIGH 20"
    assert test_com.get_written() == "SOURce:POWer:LIMit:HIGH 10000"

    # set upper limits, limit fail
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "20 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5 A")
    with pytest.raises(ValueError):
        testdev.set_upper_limits(10, 20, 10000)
    assert test_com.get_written() is None


def test_set_lower_limits(testdev: PSI9000) -> None:
    test_com = testdev.com

    # set lower limits, success
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "80 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "2000 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "2 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5.2 A")
    testdev.set_lower_limits(2, 5.2)
    assert test_com.get_written() == "SOURce:VOLTage:LIMit:LOW 2"
    assert test_com.get_written() == "SOURce:CURRent:LIMit:LOW 5.2"

    # set lower limits, set fail
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "80 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "2000 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "2 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "50 A")
    with pytest.raises(PSI9000Error):
        testdev.set_lower_limits(2, 5)
    assert test_com.get_written() == "SOURce:VOLTage:LIMit:LOW 2"
    assert test_com.get_written() == "SOURce:CURRent:LIMit:LOW 5"

    # set lower limits, limit fail
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "1 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "20 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    with pytest.raises(ValueError):
        testdev.set_lower_limits(0.5, 500)
    assert test_com.get_written() is None


def test_set_voltage_current(testdev: PSI9000) -> None:
    power_supply = testdev
    test_com = power_supply.com

    test_com.put_name("SOURce:VOLTage?", "20.2 V")
    test_com.put_name("SOURce:CURRent?", "123.4 A")
    power_supply.set_voltage_current(20.2, 123.4)
    assert test_com.get_written() == "SOURce:VOLTage 20.200000"
    assert test_com.get_written() == "SOURce:CURRent 123.400000"

    test_com.put_name("SOURce:VOLTage?", "20.2 V")
    test_com.put_name("SOURce:CURRent?", "10 A")
    test_com.put_name("SOURce:VOLTage:LIMit:HIGH?", "20 V")
    test_com.put_name("SOURce:CURRent:LIMit:HIGH?", "2000 A")
    test_com.put_name("SOURce:POWer:LIMit:HIGH?", "10000 W")
    test_com.put_name("SOURce:VOLTage:LIMit:LOW?", "1 V")
    test_com.put_name("SOURce:CURRent:LIMit:LOW?", "5 A")
    with pytest.raises(PSI9000Error):
        power_supply.set_voltage_current(20.2, 123.4)
    assert test_com.get_written() == "SOURce:VOLTage 20.200000"
    assert test_com.get_written() == "SOURce:CURRent 123.400000"


def test_check_master_slave_config(testdev: PSI9000) -> None:
    power_supply = testdev
    test_com = power_supply.com

    # MS disabled
    test_com.put_name("SYSTem:MS:ENABle?", "OFF")
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() is None

    # device is not MASTER
    test_com.put_name("SYSTem:MS:ENABle?", "ON")
    test_com.put_name("SYSTem:MS:LINK?", "SLAVE")
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() is None

    # initialization failed
    test_com.put_name("SYSTem:MS:ENABle?", "ON")
    test_com.put_name("SYSTem:MS:LINK?", "MASTER")
    test_com.put_name("SYSTem:MS:CONDition?", "NOT_INIT")
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() == "SYSTem:MS:INITialisation"

    # wrong expected voltage, current nominal values
    test_com.put_name("SYSTem:MS:ENABle?", "ON")
    test_com.put_name("SYSTem:MS:LINK?", "MASTER")
    test_com.put_name("SYSTem:MS:CONDition?", "INIT")
    test_com.put_name("SYSTem:MS:NOMinal:VOLTage?", f"{PSI9000.MS_NOMINAL_VOLTAGE} V")
    test_com.put_name(
        "SYSTem:MS:NOMinal:CURRent?", f"{PSI9000.MS_NOMINAL_CURRENT - 1} A"
    )
    with pytest.raises(PSI9000Error):
        power_supply.check_master_slave_config()
    assert test_com.get_written() == "SYSTem:MS:INITialisation"

    # all right
    # wrong expected voltage, current nominal values
    test_com.put_name("SYSTem:MS:ENABle?", "ON")
    test_com.put_name("SYSTem:MS:LINK?", "MASTER")
    test_com.put_name("SYSTem:MS:CONDition?", "INIT")
    test_com.put_name("SYSTem:MS:NOMinal:VOLTage?", f"{PSI9000.MS_NOMINAL_VOLTAGE} V")
    test_com.put_name("SYSTem:MS:NOMinal:CURRent?", f"{PSI9000.MS_NOMINAL_CURRENT} A")
    power_supply.check_master_slave_config()
    assert test_com.get_written() == "SYSTem:MS:INITialisation"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Test module for the supercube constants enums.

NOTE: After a variable is set for the first time, the type is fixed! This can usually
happen for numerical values. Therefore, we should always use 'float' as datatype for
numerical variables.

GOOD:
demo_opcua_server.set_var(constants.Power.voltage_target, 0.0)

BAD:
demo_opcua_server.set_var(constants.Power.voltage_target, 0)
"""

import pytest

from hvl_ccb.dev.cube import alarms, constants
from hvl_ccb.dev.cube.earthing_stick import _EarthingStick


def test_measurements_scaled_input() -> None:
    assert (
        constants._MeasurementChannel("", 1, 100)._CMD_SCALE
        == '"DB_Measurements"."sx_volts_input_1"'
    )
    assert (
        constants._MeasurementChannel("", 3, 100)._CMD_SCALE
        == '"DB_Measurements"."sx_volts_input_3"'
    )


def test_measurements_divider_ratio() -> None:
    assert (
        constants._MeasurementChannel("", 1, 100)._CMD_RATIO
        == '"DB_Measurements"."si_Divider_Ratio_1"'
    )
    assert (
        constants._MeasurementChannel("", 3, 100)._CMD_RATIO
        == '"DB_Measurements"."si_Divider_Ratio_3"'
    )


def test_earthing_stick_status() -> None:
    assert (
        _EarthingStick("", 3)._CMD_STATUS
        == '"DB_Safety_Circuit"."Earthstick_3"."si_HMI_Status"'
    )


def test_earthing_stick_manual() -> None:
    assert (
        _EarthingStick("", 2)._CMD_MANUAL
        == '"DB_Safety_Circuit"."Earthstick_2"."sx_earthing_manually"'
    )


def test_alarms() -> None:
    assert alarms._Alarms.alarm(33) == alarms._Alarms.ALARM_33
    for n in (0, alarms.NUMBER_OF_ALARMS):
        with pytest.raises(ValueError):
            alarms._Alarms.alarm(n)


def test_door() -> None:
    assert constants._Door(3, "")._CMD == '"DB_Safety_Circuit"."Door_3"."si_HMI_status"'


def test_alarm_text() -> None:
    assert alarms._AlarmText.get_level(1) == alarms._AlarmLevel.STOP
    assert alarms._AlarmText.get_coming_message(1) == "Emergency Stop 1 triggered"
    assert alarms._AlarmText.get_going_message(1) == "Emergency Stop 1 released"
    assert alarms._AlarmText.get_level(1000) == alarms._AlarmLevel.NOT_DEFINED
    assert alarms._AlarmText.get_coming_message(1000) == "NO ALARM TEXT DEFINED"
    assert alarms._AlarmText.get_going_message(1000) == "NO ALARM TEXT DEFINED"


def test_message_board() -> None:
    assert constants.MessageBoard.line(3) == constants.MessageBoard.LINE_3
    for n in (0, constants.NUMBER_OF_LINES):
        with pytest.raises(ValueError):
            constants.MessageBoard.line(n)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device classes for the PFA-2 filter from Precision Filter Inc..

The PFA-2 filter is a high-precision, low-noise analog filter commonly used in
laboratory instrumentation and signal conditioning setups. It is suitable for
applications requiring clean signal output, such as spectroscopy, low-level
voltage measurements, signal recovery and high-resolution data acquisition.

The class PFA2Filter has been tested with a PFA-2 (with option H) unit.
Verify compatibility with your specific filter variant before use,
especially for units with custom or user-modified settings.

Product information and technical details:
https://pfinc.com/product/precision-pfa-2-filter-amplifier
"""

from .base import (  # noqa: F401
    Pfa2FilterChannelCoupling,
    Pfa2FilterChannelMode,
    Pfa2FilterHPFState,
    Pfa2FilterLPFMode,
    Pfa2FilterOverloadMode,
    Pfa2FilterPostGain,
    Pfa2FilterPreGain,
    Pfa2FilterSerialCommunication,
    Pfa2FilterSerialCommunicationConfig,
)
from .device import (  # noqa: F401
    Pfa2Filter,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device classes for Heinzinger Digital Interface I/II and Heinzinger PNC power supply.

The Heinzinger Digital Interface I/II is used for many Heinzinger power units.
Interface Manual:
https://www.heinzinger.com/assets/uploads/downloads/Handbuch_DigitalInterface_2021-12-14-V1.6.pdf

The Heinzinger PNC series is a series of high voltage direct current power supplies.
The class Heinzinger is tested with different PNChp-types.
Check the code carefully before using it with other PNC devices, especially PNC3p
or PNCcap.
Manufacturer homepage:
https://www.heinzinger.com/en/products/pnc-serie
"""

import logging
import re
from time import sleep
from typing import cast

from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import SingleCommDevice
from hvl_ccb.dev.heinzinger.base import HeinzingerSerialCommunication
from hvl_ccb.dev.heinzinger.constants import (
    HeinzingerDeviceNotRecognizedError,
    HeinzingerSetValueError,
    RecordingsEnum,
)
from hvl_ccb.dev.heinzinger.mixin import DeprecatedHeinzingerMixin
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_bool, validate_number

logger = logging.getLogger(__name__)


@configdataclass
class HeinzingerConfig:
    """
    Device configuration dataclass for Heinzinger power supplies.
    """

    #: default number of recordings used in averaging the current
    #  or the voltage [1, 2, 4, 8, 16]
    default_number_of_recordings: int | RecordingsEnum = 1

    #: number of decimals sent for setting the current limit or the voltage, between 1
    #  and 10
    number_of_decimals: int = 6

    #: Time to wait after subsequent commands during stop (in seconds)
    wait_sec_stop_commands: Number = 0.5

    def clean_values(self) -> None:
        if not isinstance(self.default_number_of_recordings, RecordingsEnum):
            self.force_value(  # type: ignore[attr-defined]
                "default_number_of_recordings",
                RecordingsEnum(self.default_number_of_recordings),
            )

        if self.number_of_decimals not in range(1, 11):
            msg = "The number of decimals should be an integer between 1 and 10."
            raise ValueError(msg)

        if self.wait_sec_stop_commands <= 0:
            msg = (
                "Wait time after subsequent commands during stop must be be a "
                "positive value (in seconds)."
            )
            raise ValueError(msg)


class Heinzinger(SingleCommDevice, DeprecatedHeinzingerMixin):
    """
    Heinzinger power supply device class.
    The power supply is controlled over a Heinzinger Digital Interface I/II

    Sends basic SCPI commands and reads the answer.
    Only the standard instruction set from the manual is implemented.
    """

    def __init__(self, com, dev_config=None) -> None:
        # Call superclass constructor
        super().__init__(com, dev_config)

        # Version of the interface (will be retrieved after com is opened)
        self._interface_version = ""

        # Status of the voltage output (it has to be updated via the output = True and
        # output = False because querying it is not supported)
        self._output_status: bool | None = None

        # Serial number of the device (will be retrieved after com is opened)
        self._serial_number: str = ""
        # model of the device (derived from serial number)
        self._model: str = ""
        # maximum output current of the hardware (unit A)
        self._max_current_hardware: Number = 0
        # maximum output voltage of the hardware (unit V)
        self._max_voltage_hardware: Number = 0
        # maximum output current set by user (unit A)
        self._max_current: Number = 0
        # maximum output voltage set by user (unit V)
        self._max_voltage: Number = 0
        # the tolerance value for checking if the set voltage/current is correctly set
        self._epsilon: Number = 1e-3
        # identify voltage range, because unit is always V
        self._voltage_multiplier: Number = 0
        # identify current range, because unit is always A
        self._current_multiplier: Number = 0

    def __repr__(self) -> str:
        return (
            f"HeinzingerPNC({self._serial_number}), with "
            f"HeinzingerDI({self._interface_version})"
        )

    @staticmethod
    def default_com_cls() -> type[HeinzingerSerialCommunication]:
        return HeinzingerSerialCommunication

    @staticmethod
    def config_cls() -> type[HeinzingerConfig]:
        return HeinzingerConfig

    def start(self) -> None:
        """
        Opens the communication protocol and configures the device.

        :raises SerialCommunicationIOError: when communication port cannot be opened.
        """

        logger.info(f"Starting device {self}")
        super().start()

        self._interface_version = self.get_interface_version()

        # find out which type of source this is:
        self.identify_device()
        self.number_of_recordings = self.config.default_number_of_recordings

    def stop(self) -> None:
        """
        Stop the device. Closes also the communication protocol.
        """

        logger.info(f"Stopping device {self}")
        if not self.com.is_open:
            logger.warning(f"Device {self} already stopped")
        else:
            # set the voltage to zero
            self.voltage = 0
            sleep(self.config.wait_sec_stop_commands)
            # switch off the voltage output
            self.output = False
            sleep(self.config.wait_sec_stop_commands)
        super().stop()

    def identify_device(self) -> None:
        """
        Identify the device nominal voltage and current based on its serial number.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        serial_number = self.get_serial_number()
        # regex to find the model of the device
        regex_vc = r"(\d+)-(\d+)"  # voltage-current info
        regex_model = r"PNC.*?" + regex_vc + r"\s?[a-z]{3}"
        result = re.search(regex_model, serial_number)
        if result:
            self._serial_number = serial_number
            model = result.group()
            self._model = model
            # regex to find the nominal voltage and nominal current
            match = cast("re.Match", re.search(regex_vc, model))
            voltage = int(match.group(1))
            current = int(match.group(2))
            # the units of voltage (V) and current (A)
            self._max_voltage_hardware = voltage
            self._max_voltage = voltage
            self._max_current_hardware = current / 1000
            self._max_current = current / 1000
            if self._max_voltage_hardware < 100000:
                self._voltage_multiplier = 1.0
            else:
                self._voltage_multiplier = 1e3

            if self._max_current_hardware < 0.001:
                self._current_multiplier = 1e-6  # I_nenn < 1 mA
            elif self._max_current_hardware > 1:
                self._current_multiplier = 1.0  # I_nenn
            else:
                self._current_multiplier = 1e-3  # I_nenn >= 1 mA & < 1 A
            logger.info(f"Device {model} successfully identified")
        else:
            raise HeinzingerDeviceNotRecognizedError(serial_number)

    def reset_interface(self) -> None:
        """
        Reset of the digital interface; only Digital Interface I:
        Power supply is switched to the Local-Mode (Manual operation)

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        logger.info("Reset of the digital interface")
        self.com.write("*RST")

    def get_interface_version(self) -> str:
        """
        Queries the version number of the digital interface.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        version = self.com.query("VERS?")
        logger.info(f"The interface version of the Heinzinger: {version}")
        return version

    def get_serial_number(self) -> str:
        """
        Ask the device for its serial number and returns the answer as a string.

        :return: string containing the device serial number
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = self.com.query("*IDN?")
        logger.info(f"The identification string of the Heinzinger: {value}")
        return value

    @property
    def max_current_hardware(self) -> Number:
        """Maximal output current the hardware can provide in A"""
        return self._max_current_hardware

    @property
    def max_voltage_hardware(self) -> Number:
        """Maximal output voltage the hardware can provide in V"""
        return self._max_voltage_hardware

    @property
    def max_current(self) -> Number:
        """Maximal settable output current in A"""
        return self._max_current

    @max_current.setter
    def max_current(self, value: Number | None):
        """Maximal settable output current in A"""
        if value is None:
            # Reset the user-defined limit to the hardware limit
            self._max_current = self.max_current_hardware
            return
        validate_number(
            "max_current", value, (0, self.max_current_hardware), logger=logger
        )
        self._max_current = value

    @property
    def max_voltage(self) -> Number:
        """Maximal settable output voltage in V"""
        return self._max_voltage

    @max_voltage.setter
    def max_voltage(self, value: Number | None):
        """Maximal settable output voltage in V"""
        if value is None:
            # Reset the user-defined limit to the hardware limit
            self._max_voltage = self.max_voltage_hardware
            return
        validate_number(
            "max_voltage", value, (0, self.max_voltage_hardware), logger=logger
        )
        self._max_voltage = value

    @property
    def output(self) -> bool | None:
        """
        Switch DC voltage output on and updates the output status.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        return self._output_status

    @output.setter
    def output(self, value: bool) -> None:
        """
        Switch DC voltage output on or off and updates the output status.

        :raises SerialCommunicationIOError: when communication port is not opened
        :raises TypeError: if value is not a `bool`
        """
        validate_bool("Enable DC voltage output", value, logger)

        if value:
            self.com.write("OUTP ON")
            self._output_status = True
        else:
            self.com.write("OUTP OFF")
            self._output_status = False
        logger.info(f"DC voltage output is {'ON' if value else 'OFF'}")

    @property
    def number_of_recordings(self) -> int:
        """
        Queries the number of recordings the device is using for average value
        calculation.

        :return: int number of recordings
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = self.com.query("AVER?")
        logger.info(
            "the number of recordings the device is using for "
            f"average value calculation: {value}"
        )
        return int(value)

    @number_of_recordings.setter
    def number_of_recordings(self, value: int | RecordingsEnum) -> None:
        """
        Sets the number of recordings the device is using for average value
        calculation. The possible values are 1, 2, 4, 8 and 16.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = int(RecordingsEnum(value).value)
        # if the input is not 1, 2, 4, 8, or 16, RangeEnum picks a suitable one
        validate_number("number_of_recordings", value, (1, 16), int, logger=logger)
        self.com.write(f"AVER {value}")

    @property
    def voltage(self) -> Number:
        """
        Ask the Device to measure its output voltage and return the measurement
        result in V.

        :return: measured voltage as float in V
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = float(self.com.query("MEAS:VOLT?")) * self._voltage_multiplier
        logger.info(f"Output measured voltage of the Heinzinger PNC: {value} V")
        return value

    @voltage.setter
    def voltage(self, value: Number) -> None:
        """
        Sets the output voltage of the Heinzinger PNC to the given value in V.
        Same as set_voltage.setter

        :param value: voltage expressed in V
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises HeinzingerSetValueError: when the value was not set correctly
        """
        validate_number("voltage", value, (0, self.max_voltage), logger=logger)
        # value_input could be in kV or V depending on self._max_voltage_hardware
        value_command = value / self._voltage_multiplier
        self.com.write(f"VOLT {value_command:.{self.config.number_of_decimals}f}")
        # Validation:
        sleep(self.config.wait_sec_stop_commands)
        set_value = self.set_voltage
        if abs(value - set_value) > self._epsilon * self._max_voltage_hardware:
            msg = (
                f"Tried to set voltage = {value} V, "
                f"but the device did not take the value; the value is = {set_value} V"
            )
            logger.error(msg)
            raise HeinzingerSetValueError(msg)

        logger.info(f"Output voltage of the Heinzinger PNC is set to: {value} V")

    @property
    def set_voltage(self) -> Number:
        """
        Queries the set voltage of the Heinzinger PNC (not the measured voltage!) in V.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = float(self.com.query("VOLT?")) * self._voltage_multiplier
        logger.info(f"Output set voltage of the Heinzinger PNC: {value} V")
        return value

    @set_voltage.setter
    def set_voltage(self, value: Number) -> None:
        """
        Sets the output voltage of the Heinzinger PNC to the given value in V.
        Same as voltage.setter

        :param value: voltage expressed in V
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        self.voltage = value

    @property
    def current(self) -> Number:
        """
        Ask the Device to measure its output current and return the measurement
        result in A.

        :return: measured current as float
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = float(self.com.query("MEAS:CURR?")) * self._current_multiplier
        logger.info(f"Output measured current of the Heinzinger PNC: {value} A")
        return value

    @current.setter
    def current(self, value: Number) -> None:
        """
        Sets the output current of the Heinzinger PNC to the given value in A.
        Same as set_current.setter

        :param value: current expressed in A
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises HeinzingerSetValueError: when the value was not set correctly
        """
        validate_number("current", value, (0, self.max_current), logger=logger)
        # value_input could be in µA, mA, or A depending on self._max_current_hardware
        value_command = value / self._current_multiplier
        self.com.write(f"CURR {value_command:.{self.config.number_of_decimals}f}")
        # Validation:
        sleep(self.config.wait_sec_stop_commands)
        set_value = self.set_current
        if abs(value - set_value) > self._epsilon * self._max_current_hardware:
            msg = (
                f"Tried to set current = {value} A, "
                f"but the device did not take the value; the value is = {set_value} A"
            )
            logger.error(msg)
            raise HeinzingerSetValueError(msg)

        logger.info(f"Output current of the Heinzinger PNC is set to: {value} A")

    @property
    def set_current(self) -> Number:
        """
        Queries the set current of the Heinzinger PNC (not the measured current!) in A.

        :raises SerialCommunicationIOError: when communication port is not opened
        """
        value = float(self.com.query("CURR?")) * self._current_multiplier
        logger.info(f"Output set current of the Heinzinger PNC: {value} A")
        return value

    @set_current.setter
    def set_current(self, value: Number) -> None:
        """
        Sets the output current of the Heinzinger PNC to the given value in A.
        Same as current.setter

        :param value: current expressed in A
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        self.current = value



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Alarms of the different "Cubes".
"""

from aenum import IntEnum

from hvl_ccb.utils.enum import ValueEnum, unique

from .constants import _PrefixedNumbersEnumBase

NUMBER_OF_ALARMS = 152


class _AlarmEnumBase(_PrefixedNumbersEnumBase):
    """
    Base class for enums with "Alarm{n}" instance names, where n=1..N.
    """

    @classmethod
    def _prefix(cls) -> str:
        return "ALARM_"

    @classmethod
    def alarm(cls, number: int):  # noqa: ANN206
        """
        Get the enum instance for a given alarm number.

        :param number: the alarm number (1..N)
        :return: the enum instance for the alarm number.
        :raises ValueError: when alarm number is not in the 1..N range
        """
        return cls.get(number)


_Alarms = unique(
    _AlarmEnumBase(
        "Alarms",
        {
            f"{_AlarmEnumBase._prefix()}{n}": f'"DB_Alarm_HMI"."Alarm{n}"'
            for n in range(1, NUMBER_OF_ALARMS)
        },
    )
)
"""
Alarms enumeration containing all variable NodeID strings for the alarm array.
"""


class _AlarmStatus(IntEnum):
    INACTIVE = 0
    ACTIVE = 1


class _AlarmsOverview:
    """
    Stores the status of all alarms / messages
    """

    def __init__(self) -> None:
        for i in _Alarms.range():
            setattr(type(self), f"alarm_{i}", _AlarmStatus.INACTIVE)


class _AlarmLevel(IntEnum):
    """
    Alarm Level for OPC alarms, the level is similar to the corresponding logging level.
    """

    NOT_DEFINED = 10
    MESSAGE = 20
    WARNING = 30
    STOP = 35


class _AlarmText(ValueEnum, init="level coming going"):  # type: ignore[call-arg]
    """
    This enumeration contains the message for coming and going alarms of the BaseCube
    system.
    The corresponding AlarmLevel is also stored.
    Use the :meth:`AlarmText.get_level`
    method to retrieve the alarm level of an alarm number.
    Use the :meth:`AlarmText.get_coming_message`
    method to retrieve the going message of an alarm number.
    Use the :meth:`AlarmText.get_going_message`
    method to retrieve the coming message of an alarm number.
    """

    # Safety elements
    ALARM_1 = (
        _AlarmLevel.STOP,
        "Emergency Stop 1 triggered",
        "Emergency Stop 1 released",
    )
    ALARM_2 = (
        _AlarmLevel.STOP,
        "Emergency Stop 2 triggered",
        "Emergency Stop 2 released",
    )
    ALARM_3 = (
        _AlarmLevel.STOP,
        "Emergency Stop 3 triggered",
        "Emergency Stop 3 released",
    )
    ALARM_4 = _AlarmLevel.STOP, "Safety Switch 1 Error", "Safety Switch 1 Error solved"
    ALARM_5 = _AlarmLevel.STOP, "Safety Switch 2 Error", "Safety Switch 2 Error solved"
    ALARM_6 = _AlarmLevel.STOP, "Door 1 Lock Failure", "Door 1 Lock Failure solved"
    ALARM_7 = _AlarmLevel.STOP, "Door 2 Lock Failure", "Door 2 Lock Failure solved"
    ALARM_8 = _AlarmLevel.STOP, "Door 3 Lock Failure", "Door 3 Lock Failure solved"
    ALARM_9 = (
        _AlarmLevel.STOP,
        "Earthing Stick 1 Error while opening",
        "Earthing Stick 1 Error while opening solved",
    )
    ALARM_10 = (
        _AlarmLevel.STOP,
        "Earthing Stick 2 Error while opening",
        "Earthing Stick 2 Error while opening solved",
    )
    ALARM_11 = (
        _AlarmLevel.STOP,
        "Earthing Stick 3 Error while opening",
        "Earthing Stick 3 Error while opening solved",
    )
    ALARM_12 = (
        _AlarmLevel.STOP,
        "Earthing Stick 4 Error while opening",
        "Earthing Stick 4 Error while opening solved",
    )
    ALARM_13 = (
        _AlarmLevel.STOP,
        "Earthing Stick 5 Error while opening",
        "Earthing Stick 5 Error while opening solved",
    )
    ALARM_14 = (
        _AlarmLevel.STOP,
        "Earthing Stick 6 Error while opening",
        "Earthing Stick 6 Error while opening solved",
    )
    ALARM_15 = (
        _AlarmLevel.STOP,
        "Earthing Stick 1 Error while closing",
        "Earthing Stick 1 Error while closing solved",
    )
    ALARM_16 = (
        _AlarmLevel.STOP,
        "Earthing Stick 2 Error while closing",
        "Earthing Stick 2 Error while closing solved",
    )
    ALARM_17 = (
        _AlarmLevel.STOP,
        "Earthing Stick 3 Error while closing",
        "Earthing Stick 3 Error while closing solved",
    )
    ALARM_18 = (
        _AlarmLevel.STOP,
        "Earthing Stick 4 Error while closing",
        "Earthing Stick 4 Error while closing solved",
    )
    ALARM_19 = (
        _AlarmLevel.STOP,
        "Earthing Stick 5 Error while closing",
        "Earthing Stick 5 Error while closing solved",
    )
    ALARM_20 = (
        _AlarmLevel.STOP,
        "Earthing Stick 6 Error while closing",
        "Earthing Stick 6 Error while closing solved",
    )
    ALARM_21 = _AlarmLevel.STOP, "Safety Fence 1 not closed", "Safety Fence 1 closed"
    ALARM_22 = _AlarmLevel.STOP, "Safety Fence 2 not closed", "Safety Fence 2 closed"
    ALARM_23 = _AlarmLevel.STOP, "OPC Connection Error", "OPC Connection Error is gone"
    ALARM_24 = _AlarmLevel.STOP, "Grid Power Failure", "Grid Power Failure is gone"
    ALARM_25 = _AlarmLevel.STOP, "UPS Failure", "UPS Failure is gone"
    ALARM_26 = _AlarmLevel.STOP, "24V PSU Failure", "24V PSU Failure is gone"

    # Power unit
    ALARM_27 = (
        _AlarmLevel.STOP,
        "Power Setup and Power Switch Position are not matching",
        "Power Setup and Power Switch Position are matching again",
    )
    ALARM_28 = (
        _AlarmLevel.STOP,
        "Power Inverter Failure",
        "Power Inverter Failure solved",
    )
    ALARM_29 = (
        _AlarmLevel.STOP,
        "Control Loop Response Failure",
        "Control Loop Response Failure solved",
    )
    ALARM_30 = (
        _AlarmLevel.STOP,
        "'Set Polarity' does not match with Measured Voltage",
        "'Set Polarity' matches with Measured Voltage",
    )

    # Doors
    ALARM_41 = (
        _AlarmLevel.WARNING,
        "Door 1: Use Earthing Rod!",
        "Door 1: Earthing Rod used.",
    )
    ALARM_42 = (
        _AlarmLevel.MESSAGE,
        "Door 1: Earthing Rod is still in Experiment.",
        "Door 1: Earthing Rod is removed from Experiment.",
    )
    ALARM_43 = (
        _AlarmLevel.WARNING,
        "Door 2: Use Earthing Rod!",
        "Door 2: Earthing Rod used.",
    )
    ALARM_44 = (
        _AlarmLevel.MESSAGE,
        "Door 2: Earthing Rod is still in Experiment.",
        "Door 2: Earthing Rod is removed from Experiment.",
    )
    ALARM_45 = (
        _AlarmLevel.WARNING,
        "Door 3: Use Earthing Rod!",
        "Door 3: Earthing Rod used.",
    )
    ALARM_46 = (
        _AlarmLevel.MESSAGE,
        "Door 3: Earthing Rod is still in Experiment.",
        "Door 3: Earthing Rod is removed from Experiment.",
    )

    # General
    ALARM_47 = _AlarmLevel.MESSAGE, "UPS Charge < 85%", "UPS Charge >= 85%"
    ALARM_48 = _AlarmLevel.MESSAGE, "UPS running on Battery", "UPS running on Grid"
    ALARM_49 = (
        _AlarmLevel.WARNING,
        "Remove PD-Calibrator from the Circuit",
        "PD-Calibrator removed from the Circuit",
    )
    ALARM_50 = _AlarmLevel.MESSAGE, "OPC Connection active", "OPC Connection not active"
    ALARM_57 = (
        _AlarmLevel.MESSAGE,
        "Breakdown Detection Unit triggered",
        "Breakdown Detection Unit is reset",
    )

    # Earthing Stick
    ALARM_51 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 1: Manual earthing enabled",
        "Earthing Stick 1: Manual earthing disabled",
    )
    ALARM_52 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 2: Manual earthing enabled",
        "Earthing Stick 2: Manual earthing disabled",
    )
    ALARM_53 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 3: Manual earthing enabled",
        "Earthing Stick 3: Manual earthing disabled",
    )
    ALARM_54 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 4: Manual earthing enabled",
        "Earthing Stick 4: Manual earthing disabled",
    )
    ALARM_55 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 5: Manual earthing enabled",
        "Earthing Stick 5: Manual earthing disabled",
    )
    ALARM_56 = (
        _AlarmLevel.MESSAGE,
        "Earthing Stick 6: Manual earthing enabled",
        "Earthing Stick 6: Manual earthing disabled",
    )

    # generic not defined alarm text
    NOT_DEFINED = (
        _AlarmLevel.NOT_DEFINED,
        "NO ALARM TEXT DEFINED",
        "NO ALARM TEXT DEFINED",
    )

    @classmethod
    def get_level(cls, alarm: int) -> _AlarmLevel:
        """
        Get the alarm level of this enum for an alarm number.

        :param alarm: the alarm number
        :return: the alarm level for the desired alarm number
        """

        try:
            return getattr(cls, f"ALARM_{alarm}").level
        except AttributeError:
            return cls.NOT_DEFINED.level  # type: ignore[attr-defined]

    @classmethod
    def get_coming_message(cls, alarm: int) -> str:
        """
        Get the coming message of this enum for an alarm number.

        :param alarm: the alarm number
        :return: the coming alarm message for the desired alarm number
        """

        try:
            return getattr(cls, f"ALARM_{alarm}").coming
        except AttributeError:
            return cls.NOT_DEFINED.coming  # type: ignore[attr-defined]

    @classmethod
    def get_going_message(cls, alarm: int) -> str:
        """
        Get the going message of this enum for an alarm number.

        :param alarm: the alarm number
        :return: the going alarm message for the desired alarm number
        """

        try:
            return getattr(cls, f"ALARM_{alarm}").going
        except AttributeError:
            return cls.NOT_DEFINED.going  # type: ignore[attr-defined]



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Base classes for the PFA-2 filter from Precision Filter Inc..

The PFA-2 filter is a high-precision, low-noise analog filter commonly used in
laboratory instrumentation and signal conditioning setups. It is suitable for
applications requiring clean signal output, such as spectroscopy, low-level
voltage measurements, signal recovery and high-resolution data acquisition.

The class PFA2Filter has been tested with a PFA-2 (with option H) unit.
Verify compatibility with your specific filter variant before use,
especially for units with custom or user-modified settings.

Product information and technical details:
https://pfinc.com/product/precision-pfa-2-filter-amplifier
"""

from typing import Literal

from aenum import StrEnum

from hvl_ccb.comm.base import CommunicationError, SyncCommunicationProtocol
from hvl_ccb.comm.serial import SerialCommunication, SerialCommunicationConfig
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError
from hvl_ccb.utils.enum import RangeEnum

N_CHANNELS: int = 2
LPF_FREQUENCY_MIN: int = 5
LPF_FREQUENCY_MAX: int = 127_750
HPF_FREQUENCY_MIN: int = 1
HPF_FREQUENCY_MAX: int = 127_500


class Pfa2FilterCommunicationError(CommunicationError):
    pass


class Pfa2FilterError(DeviceError):
    pass


@configdataclass
class Pfa2FilterSerialCommunicationConfig(SerialCommunicationConfig):
    """
    Specific communication config implementation for for PFA-2 filters
    Predefines device-specific protocol parameters.
    """

    baudrate: int = 57600  # PFA-2 filter uses 57600 baud


class Pfa2FilterSerialCommunication(SerialCommunication, SyncCommunicationProtocol):
    """
    Specific communication protocol implementation for PFA-2 filters.
    Predefines device-specific protocol parameters in config.
    """

    def __init__(self, configuration) -> None:
        super().__init__(configuration)

    @staticmethod
    def config_cls():
        return Pfa2FilterSerialCommunicationConfig


class Pfa2FilterChannelMode(StrEnum):
    """StrEnum for the channel modes"""

    OPERATE = "OPERATE"
    CALIBRATION = "CAL"
    SHORTED = "SHORT"


class Pfa2FilterChannelCoupling(StrEnum):
    """StrEnum for the channel coupling"""

    AC = "AC"
    DC = "DC"


class Pfa2FilterLPFMode(StrEnum):
    """StrEnum for the low pass filter mode"""

    FLAT = "FLAT"
    PULSE = "PULSE"
    OFF = "NONE"


class Pfa2FilterHPFState(StrEnum):
    """StrEnum for the high pass filter state"""

    ON = "ON"
    OFF = "OFF"


class Pfa2FilterPreGain(RangeEnum):
    """RangeEnum for the pre filter gain value"""

    @classmethod
    def unit(cls) -> Literal[""]:
        return ""

    @classmethod
    def is_reversed(cls) -> bool:
        return True

    ONE: float = 1
    TWO: float = 2
    FOUR: float = 4
    EIGHT: float = 8
    SIXTEEN: float = 16
    THIRTY_TWO: float = 32
    SIXTY_FOUR: float = 64
    ONE_HUNDRED_TWENTY_EIGHT: float = 128


class Pfa2FilterPostGain(RangeEnum):
    """RangeEnum for the post filter gain value"""

    @classmethod
    def unit(cls) -> Literal[""]:
        return ""

    @classmethod
    def is_reversed(cls) -> bool:
        return True

    ONE_SIXTEENTH: float = 0.0625
    ONE_EIGHT: float = 0.125
    ONE_QUARTER: float = 0.25
    ONE_HALF: float = 0.5
    ONE: float = 1
    TWO: float = 2
    FOUR: float = 4
    EIGHT: float = 8
    SIXTEEN: float = 16


class Pfa2FilterOverloadMode(StrEnum):
    """StrEnum for the overload handling mode"""

    LATCHING = "LATCHING"
    CONTINUOUS = "CONTINUOUS"


class _Pfa2FilterCommands(StrEnum):
    """StrEnum for the PFA-2 filter commands"""

    COUPLING = "COUPLING"
    MODE = "MODE"
    PREGAIN = "PREGAIN"
    POSTGAIN = "POSTGAIN"
    LPF_TYPE = "LPFILTTYPE"
    LPF_FREQUENCY = "LPFC"
    HPF_STATE = "HPFILT"
    HPF_TYPE = "HPFILTTYPE"
    HPF_FREQUENCY = "HPFC"
    IN_OVERLOAD = "INOVLD"
    OUT_OVERLOAD = "OUTOVLD"
    OVERLOAD_LIMIT = "OUTOVLDLIM"
    OVERLOAD_MODE = "OVLDMODE"
    OVERLOAD_CLEAR = "OVLDCLEAR"

    def build_str(self, channel: int, param: str | None = None) -> str:
        """
        Build a command string for sending to the device.

        getter example: Pfa2FilterCommands.COUPLING.build_str(1)-->"1:COUPLING?"
        setter example: Pfa2FilterCommands.COUPLING.build_str(1, 'AC')-->"1:COUPLING=AC"


        :param param: Command's parameter given as string
        :return: Command's string
        """
        action = "?" if param is None else f"={param}"

        return f"{channel}:{self.value}{action}"



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
DC Laboratory Power Supply Unit "Korad KA3000P"

Interface for the power supply units of "Korad" and their "KA3000P"-series.
This package was developed and tested with "Korad KA3005P".
It might also work with other devices like the ones of the "Korad KA6000P"-series.
"""

from .base import KA3000PError as KA3000PError
from .device import KA3000P as KA3000P
